/*
 *	Client routines for Network News Tranfer Protocol ala RFC977
 *
 *	Copyright 1990 Anders Klemets - SM0RGV, All Rights Reserved.
 *	Permission granted for non-commercial copying and use, provided
 *	this notice is retained.
 *
 *	Changes copyright 1990 Bernie Roehl, All Rights Reserved.
 *	Permission granted for non-commercial copying and use, provided
 *	this notice is retained.
 *
 *  Revision history:
 *
 *     May 11, 1990 - br checked for invalid chars in news filenames
 *
 *     May 10, 1990 - br changed date stamp in 'From ' lines to
 *            seconds since GMT (to make parsing and expiry easier)
 *
 *     May 9, 1990 - br added locking of nntp.dat and history files,
 *            second parameter to NNTP DIR, fixed bug in updating of
 *            nntp.dat
 *
 *     early May, 1990 -- br added NNTP TRACE, NNTP DIR,
 *            server-specific newsgroups and connection windows,
 *            locking of newsgroup files using mlock() and rmlock(),
 *            date stamping of 'From ' lines, increased stack space,
 *            updating of nntp.dat only on successful sessions.
 *
 *     July 19, 1990 pa0gri Delinted and cleaned up. (calls and includes)
 *
 *	20 May 92	1.2		GT	Output file in "snews" batch format.
 *	21 May 92	1.3		GT	Fix 1.2.
 *	07 Jun 92	1.4		GT	"connect_wait_val" timeout on connect () call.
 *	11 Jun 92	1.5		GT	Booked out in error.
 *	15 Jun 92	1.6		GT	Flush history file write.
 *
 *	21-Jul-92	Chris Sowden (csowden@mouse.demon.co.uk)
 *					Added NNTP BATCH and NNTP VERBOSE commands.
 *					Re-wrote article fetching section to allow overlapped
 *					operation (batch mode).
 *					Added a 64K bit history hash table.
 *					Added 2 mins time adjustment to allow for a fast clock.
 *					Added status message showing articles, bytes and rate.
 *					Lock "nntp.dat" for duration of session to prevent races.
 *					Numerous changes to speed things up.
 *
 *	08 Aug 92	1.8		MM	Force disc writes of history file.
 *
 *	11-Sep-92	Chris Sowden (csowden@mouse.demon.co.uk)
 *					Look for newsgroup list in data file first.
 *					Allow multiple NEWNEWS requests.
 *					Added NNTP NEWGROUPS command.
 *					Write directly to batch file in newsbatch mode.
 *					Additional messages in verbose mode.
 *					Remove zero length files.
 *
 *	14 Sep 92	GT	Use GMT form of NEWNEWS and NEWGROUPS commands.
 *
 *	14-Sep-92	Chris Sowden (csowden@mouse.demon.co.uk)
 *					Added counting to verbose received mesage.
 *
 *	15 Sep 92	Chris Sowden (csowden@mouse.demon.co.uk)
 *					Fixed reentrancy problem.
 *
 *	15 Sep 92	GT	Fixed race condition on GMT vs. local time reading
 *					between news and mail.
 *					
 * 13-Nov-92	Chris Sowden (csowden@mouse.demon.co.uk)
 *					Added SAFETY command, expanded BATCH command
 *					Uses child processes for initialisation and requests
 *					New KILL file and GET file for selective article fetching
 *					Uses correct "#! rnews nnnn" style newsbatch separator
 *					Improved history file handling (especially for duplicates)
 *					Added progress messages/statistics in verbose mode
 *					Now logs all error messages and news summary
 *					
 * 01-Dec-92	Chris Sowden (csowden@mouse.demon.co.uk)
 *					Improved the opening sequence of KILL and GET files
 *					Corrected error levels of file opens
 *					Close temporary new message file
 *					Tried to reduce likelihood of lock files being left around
 *					Removed extra linefeed from batch files
 *
 *	08 May 93	1.13	GT	Fix warnings.
 *							Reinstate 2 minute fudge factor.
 *							Drop the word "article" from the "news
 *							arrived" message.
 *							Fix signed variables problem.
 *
 *	07 Nov 93	1.14	GT	Display NEWNEWS results when "nntp verb on".
 *							Fix dot transparency.
 *							Fix blank lines in error log.
 *
 *	   May 97	1.15	GeK	
 *	   Aug 97	1.16	GeK	Added 27th kill list for folded continuation lines
 *	         	    	   	Added simple authentication support
 *	         	    	   	(AUTHINFO USER/PASS)
 *	         	    	   	Newsgroup list held in dynamic buffer
 *	28 Aug 01	1.17	SCT	Fix log() printf format bug in error()
 *	14 Jan 02	1.18	PJD	Initial nntp POST implementation
 *	07 Feb 02	1.19	SCT	Tidied, improved POST tracing, added "nntp post"
 *							subcommands
 *	09 Mar 02	1.20	SCT	POST: improved exit checks, alarm when waiting
 *							for responses, tweaked display messages, don't
 *							give up on download just because posting fails
 *	08 Apr 02	1.21	SCT	Report all POST responses, not just those < 400
 $Id: nntpcli.c 1.21 2002/04/08 10:06:14 Simon Exp $
 */

#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <sys/timeb.h>
#include <ctype.h>
#include <string.h>			 /* for strchr() */
#include <io.h>
#ifdef MSDOS
#include <dir.h>
#include <dos.h>
#endif
#include "global.h"
#include "timer.h"
#include "cmdparse.h"
#include "commands.h"
#include "socket.h"
#include "usock.h"
#include "netuser.h"
#include "proc.h"
#include "smtp.h"
#include "files.h"
#include "main.h"			/*  for main_exit  */
#include "session.h"		/*  for NNTP_SESSION  */

#undef	NNTP_SESSION		/*  GeK - runs in own session  */
#define	NNTP_RETRY			/*  GeK - retries on server busy  */
#undef	NNTP_MODE_READER	/*  GeK - sends NNTP command 'MODE READER'  */
#define	NNTP_AUTH			/*  GeK - simple AUTHINFO USER, AUTHINFO PASS support  */

/* The grouplist must fit on a line like "NEWNEWS <grouplist> 000000 000000 GMT\n" */

#define NNTPMAXLEN	512
#define NNTPMAXGROUPLIST (NNTPMAXLEN-28-1)	/*  GeK - remove 1 for INN 512 char \r\n bug  */

/*  GeK - define max Message-ID length including <> and trailing \n\0
 *  (from so1036 - INN has a limit of 255 anyway (dbz key))
 *	saves 256 bytes per nntprequest struct
 */
#define MAX_MESSAGE_ID_LEN	256

/*  GeK - made this configurable here  */
#define NEWNEWS_FUZZ	300		/*  5 minutes  */

/*  GeK - inspired by r$ and INN, assumes a and b at least 2 chrs + \0 long  */
#define FAST_STREQ(a,b)	((a)[0]==(b)[0] && (a)[1]==(b)[1] && (a)[2]==(b)[2] && strcmp ((a),(b)) == 0)

/*  GeK - replaces (strcmp(buf, ".\n") == 0) to speed things up a bit  */
#define CHECK_FOR_DOT(s)	((s)[0] == '.' && (s)[1] == '\n' && (s)[2] == '\0')

struct nntpservers	/*  GeK - removed static - this is a struct declare!  */
	{
	struct timer nntpcli_t;
	char *name;
	char *groups;
	int lowtime, hightime;	 /* for connect window */
#ifdef	NNTP_AUTH	/*  GeK - simple AUTHINFO support  */
	char *username;
	char *password;
#endif	/* NNTP_AUTH */
	struct nntpservers *next;
	};

#define	NULLNNTP	(struct nntpservers *)NULL

#define NEWSBATCH TRUE		 /* output for batching newsreader */
#define NEWSBATCHFORMAT "#! rnews %010lu\n"

#if !NEWSBATCH
#define MAXGROUPDIRS 10
static struct grouploc
	{
	char *prefix;			 /* e.g. comp, rec, net, talk, alt ... */
	char *directory;		 /* directory where these groups should be */
	}        groupdirs[MAXGROUPDIRS] =
	{
	         NULL, NULL
	};

#endif

struct nntpservers *Nntpservers = NULLNNTP;
static char *Nntpgroups = NULLCHAR;
static unsigned short nntptrace = 1;
static int nntpbatch = 0;
static int nntpbatchsize = 2;
static int nntpnewgroups = 0;
static int nntpsafety = 1;
static int nntpverbose = 1;

#ifdef	NNTP_RETRY	/*  GeK  */
static unsigned int	nntp_max_retries = 0;		/*  Max kick attempts; default = 0 (off)  */
static unsigned int	nntp_retry_delay = 0;		/*  Seconds between kicks; default = 0  */
#endif

#if	!NEWSBATCH
static char *validchars = "abcdefghijklmnopqrstuvwxyz0123456789-_";
#endif

/*  GeK - incomplete forward struct declarations to fix gcc warnings  */
struct textline;
struct nntprequest;

int ext_dokicks (void);		/*  GeK - added int to match function  */
static void nntptick __ARGS ((void *tp));

static int dobatch __ARGS ((int argc, char *argv[], void *p));
static int doadds __ARGS ((int argc, char *argv[], void *p));
static int dondir __ARGS ((int argc, char *argv[], void *p));
static int dodrops __ARGS ((int argc, char *argv[], void *p));
static int dogroups __ARGS ((int argc, char *argv[], void *p));
static int dokicks __ARGS ((int argc, char *argv[], void *p));
static int dolists __ARGS ((int argc, char *argv[], void *p));
static int donewgroups __ARGS ((int argc, char *argv[], void *p));
static int dosafety __ARGS ((int argc, char *argv[], void *p));
static int donntrace __ARGS ((int argc, char *argv[], void *p));
static int doverbose __ARGS ((int argc, char *argv[], void *p));

#ifdef	NNTP_RETRY	/*  GeK  */
static int doretry __ARGS ((int argc, char *argv[], void *p));
#endif

static void error __ARGS ((int level, char *text));

static void freetextlinelist __ARGS ((struct textline ** list));

static void createkilllists __ARGS ((void));
static void scankilllists __ARGS ((char *line, int *status));

static unsigned int hashmsgid __ARGS ((char *msgid, unsigned int modulus));
static void createhashtable __ARGS ((void));
static int isinhashtable __ARGS ((char *msgid));
static int isinhistory __ARGS ((char *msgid));
static void addtohashtable __ARGS ((char *msgid));
static void addtohistory __ARGS ((char *msgid));

static int stripgroups __ARGS ((char *s));
static unsigned long unitspersecond __ARGS ((unsigned long units, unsigned long milliseconds));

static int getreply_in_buf (char *);
static int getreply __ARGS ((void));
static int gettxt __ARGS ((FILE * fp, struct nntprequest * request, int newnews));
static int putarticle __ARGS ((FILE * msgf, long int *msgfpos, struct nntprequest * request));
static int getheader __ARGS ((struct nntprequest * request));
static void moveheadertofile __ARGS ((FILE * fp, struct nntprequest * request));

static void nntp_init __ARGS ((int argc, void *argv, void *parentproc));
static void nntp_send __ARGS ((int argc, void *argv, void *parentproc));
static void nntp_main __ARGS ((int i1, void *tp, void *v1));
#ifdef	notdef		/*  GeK - disabled this  */
static void nntp_reg __ARGS ((void));
#endif
static void internal_receive_error (int response, char *response_buf, struct nntprequest *request);

/*  GeK/PDL - directp NNTP posting  */
#ifdef	NNTP_POST
static int nntppost = 1;
static long nntppost_respwait = 60000L;		/* wait 5 mins for responses */
static int do_nntp_post (int argc, char *argv[], void *p);
static int nntp_post (void);
static FILE *validate_article (const char *fname, char **pmsgid);
static int post_article (FILE *fp, const char *msgid, const char *fname);
static char *skipws (const char *s);
#endif

/* Tracing levels:
	0 - no tracing
	1 - serious errors reported
	2 - transient errors reported
	3 - session progress reported
	4 - actual received articles displayed
 */

static struct cmds Nntpcmds[] = {
								 {"addserver", doadds, 0, 3,
								  "nntp addserver <nntpserver> <interval>"},
								 {"batch", dobatch, 0, 0, NULLCHAR},
								 {"directory", dondir, 0, 0, NULLCHAR},
								 {"dropserver", dodrops, 0, 2,
								  "nntp dropserver <nntpserver>"},
								 {"groups", dogroups, 0, 0, NULLCHAR},
								 {"kick", dokicks, 0, 0,
								  "nntp kick <nntpserver>"},
								 {"listservers", dolists, 0, 0, NULLCHAR},
								 {"newgroups", donewgroups, 0, 0, NULLCHAR},
#ifdef	NNTP_POST
								 {"post", do_nntp_post, 0, 0, NULLCHAR},
#endif
#ifdef	NNTP_RETRY
								 {"retry", doretry, 0, 0, NULLCHAR},
#endif
								 {"safety", dosafety, 0, 0, NULLCHAR},
								 {"trace", donntrace, 0, 0, NULLCHAR},
								 {"verbose", doverbose, 0, 0, NULLCHAR},
								 {NULLCHAR},
	};

int
    donntp (argc, argv, p)
int argc;
char *argv[];
void *p;
	{
	return subcmd (Nntpcmds, argc, argv, p);
	}

static int
    doadds (argc, argv, p)
int argc;
char *argv[];
void *p;
	{
	struct nntpservers *np;

	for (np = Nntpservers; np != NULLNNTP; np = np->next)
		if (stricmp (np->name, argv[1]) == 0)
			break;

	if (np == NULLNNTP)
		{
		np = (struct nntpservers *) callocw (1, sizeof (struct nntpservers));
		np->name = strdup (argv[1]);
		np->next = Nntpservers;
		Nntpservers = np;
		np->groups = NULLCHAR;
		np->lowtime = np->hightime = -1;
		np->nntpcli_t.func = nntptick;			 /* what to call on timeout */
		np->nntpcli_t.arg = (void *) np;
		}

	if (argc > 3)
		{
		int i;

#ifdef	notdef	/*  GeK - moved below  */
		if (np->groups == NULLCHAR)
			{
			np->groups = mallocw (NNTPMAXLEN);
			*np->groups = '\0';
			}
#endif	/* notdef */

		for (i = 3; i < argc; ++i)
			{
#ifdef	NNTP_AUTH
			if (stricmp (argv[i], "user") == 0)
				{
				i++;	/*  skip 'user'  */
				/*  reset any existing user/pass  */
				free (np->username);
				np->username = NULL;
				free (np->password);
				np->password = NULL;
				
				/*  Check for both user and passwd present  */
				if ((i+1) < argc)
					{
					np->username = strdup (argv[i++]);
					np->password = strdup (argv[i]);
					}
				}
			else
#endif	/* NNTP_AUTH */
			if (isdigit (*argv[i]))
				{
				int lh, ll, hh, hl;

				sscanf (argv[i], "%d:%d-%d:%d", &lh, &ll, &hh, &hl);
				np->lowtime = lh * 100 + ll;
				np->hightime = hh * 100 + hl;
				}
			else


			if (np->groups != NULLCHAR && 
			    (strlen (np->groups) + 1 + strlen (argv[i])) > NNTPMAXGROUPLIST)
				tprintf ("Group list too long!  Group '%s' ignored!\n", argv[i]);
			else
				{								 /* it's a group, and it
												  * fits... add it to list */
				/*  GeK - moved from above  */
				if (np->groups == NULLCHAR)
					{
					np->groups = mallocw (NNTPMAXLEN);
					*np->groups = '\0';
					}

				if (*np->groups != '\0')
					strcat (np->groups, ",");

				strcat (np->groups, argv[i]);
				}
			}

#ifdef	notdef	/*  GeK - shouldn't happen now  */
		if (*np->groups == '\0')
			{									 /* No groups specified? */
			free (np->groups);
			np->groups = NULLCHAR;
			}
#endif	/* notdef */

		}

	/* set timer duration */

	set_timer (&np->nntpcli_t, atol (argv[2]) * 1000L);
	start_timer (&np->nntpcli_t);				 /* and fire it up */
	return 0;
	}

static int
    dodrops (argc, argv, p)
int argc;
char *argv[];
void *p;
	{
	struct nntpservers *np, *npprev = NULLNNTP;

	for (np = Nntpservers; np != NULLNNTP; npprev = np, np = np->next)
		if (stricmp (np->name, argv[1]) == 0)
			{
			/*
			 *  GeK - from Niel Kempson
			 *  Don't drop the server if it's currently active
			 *  (timer stopped)
			 */
			if (!run_timer (&np->nntpcli_t))
				{
				tputs ("Unable to drop server, NNTP session active\n");
				return 0;
				}

			stop_timer (&np->nntpcli_t);
			free (np->name);
			if (np->groups)
				free (np->groups);

#ifdef	NNTP_AUTH	/*  GeK  */
			free (np->username);
			free (np->password);
#endif	/* NNTP_AUTH */

			if (npprev != NULLNNTP)
				npprev->next = np->next;
			else
				Nntpservers = np->next;

			free ((char *) np);
			return 0;
			}

	tprintf ("No such server enabled\n");
	return 0;
	}

static int
    dolists (argc, argv, p)
int argc;
char *argv[];
void *p;
	{
	struct nntpservers *np;

	for (np = Nntpservers; np != NULLNNTP; np = np->next)
		{
		char tbuf[80];

		if (np->lowtime != -1 && np->hightime != -1)
			sprintf (tbuf, " -- %02d:%02d-%02d:%02d", np->lowtime / 100, np->lowtime % 100, np->hightime / 100, np->hightime % 100);
		else
			tbuf[0] = '\0';

#if	defined (NNTP_AUTH)

		tprintf ("%-32s (%lu/%lu%s) ", np->name,
		         read_timer (&np->nntpcli_t) / 1000L,
		         dur_timer (&np->nntpcli_t) / 1000L,
		         tbuf);
		if (np->username != NULL)
			tprintf ("[%s/%s] ", np->username, np->password);
		tprintf ("%s\n", np->groups ? np->groups : "");

#else	/* !(NNTP_AUTH) */

		tprintf ("%-32s (%lu/%lu%s) %s\n", np->name,
				 read_timer (&np->nntpcli_t) / 1000L,
				 dur_timer (&np->nntpcli_t) / 1000L,
				 tbuf, np->groups ? np->groups : "");

#endif	/* !(NNTP_AUTH) */
		}

	return 0;
	}

static int donntrace (argc, argv, p)
int argc;
char *argv[];
void *p;
	{
	return setshort (&nntptrace, "NNTP tracing", argc, argv);
	}

static char *News_spool = NULL;
static int np_all = 0;		 /* non-zero if Newsdir is a malloc'ed space */

static int dondir (argc, argv, p)
int argc;
char *argv[];
void *p;
	{
	if (argc < 2)
		{
#if !NEWSBATCH
		int i;

#endif
		tprintf ("spool: %s\n", News_spool ? News_spool : Mailspool);
		tprintf ("control: %s\n", Newsdir);
#if !NEWSBATCH
		for (i = 0; i < MAXGROUPDIRS; ++i)
			if (groupdirs[i].prefix)
				tprintf ("%-10.10s %s\n", groupdirs[i].prefix, groupdirs[i].directory);
#endif

		}
	else
		{
#ifdef	notdef	/*  GeK - stripping unused code  */
#if	NEWSBATCH
		if (strchr (argv[1], '=') != NULLCHAR)
			{									 /* set a groupdir */
#endif
#if !NEWSBATCH
			char *p;

			if ((p = strchr (argv[1], '=')) != NULLCHAR)
				{								 /* set a groupdir */
				int i;

				*p++ = '\0';
				for (i = 0; i < MAXGROUPDIRS; ++i)
					if (groupdirs[i].prefix)
						if (!strnicmp (groupdirs[i].prefix, argv[1], strlen (argv[1])))
							{
							if (groupdirs[i].directory)
								{
								free (groupdirs[i].directory);
								groupdirs[i].directory = NULLCHAR;
								}

							if (*p == '\0')
								{
								free (groupdirs[i].prefix);
								groupdirs[i].prefix = NULLCHAR;
								}
							else
								groupdirs[i].directory = strdup (p);

							return 0;
							}

				if (*p == '\0')					 /* trashing a group that's
												  * not there */
					return 0;

				for (i = 0; i < MAXGROUPDIRS; ++i)
					{
					if (groupdirs[i].prefix == NULLCHAR)
						{
						groupdirs[i].prefix = strdup (argv[1]);
						if (groupdirs[i].directory)
							{
							free (groupdirs[i].directory);
							groupdirs[i].directory = NULL;
							}

						groupdirs[i].directory = strdup (p);
						return 0;
						}

					}
				tprintf ("Directory table full\n");
#else
			tprintf ("Directory table not supported\n");
#endif
			}
		else
#endif	/* notdef */	/*  GeK  */
			{									 /* no '=', so just set
												  * default */
			if (News_spool)
				free (News_spool);

			News_spool = strdup (argv[1]);
			}

		if (argc > 2)
			{									 /* they specified a newsdir
												  * as well */
			if (np_all)
				free (Newsdir);

			Newsdir = strdup (argv[2]);
			np_all = 1;
			}

		}
	return 0;
	}

int
    ext_dokicks (void)
	{
	char *args[] = {""};

	dokicks (1, args, NULL);
	return 0;
	}

static int
    dokicks (argc, argv, p)
int argc;
char *argv[];
void *p;
	{
	struct nntpservers *np;

	for (np = Nntpservers; np != NULLNNTP; np = np->next)
		if (argc == 1 || stricmp (np->name, argv[1]) == 0)
			{
			/* If the timer is not running, the timeout function has
			 * already been called and we don't want to call it again. */
			if (run_timer (&np->nntpcli_t))
				{
				stop_timer (&np->nntpcli_t);
				nntptick ((void *) np);
				}

			if (argc != 1)
				return 0;

			}

	if (argc != 1)
		tprintf ("No such server enabled.\n");

	return 0;
	}

static int
    dogroups (argc, argv, p)
int argc;
char *argv[];
void *p;
	{
	char grouplist[NNTPMAXLEN];
	int i;

	if (argc < 2)
		{
		if (Nntpgroups == NULLCHAR || (Nntpgroups != NULLCHAR && strcmp (Nntpgroups, "*") == 0))
			tprintf ("All groups are currently enabled\n");
		else
			tprintf ("Currently enabled newsgroups:\n%s\n", Nntpgroups);

		return 0;
		}

	if (Nntpgroups != NULLCHAR)
		free (Nntpgroups);

	*grouplist = '\0';
	for (i = 1; i < argc; ++i)
		{
		if ((strlen (grouplist) + 1 + strlen (argv[i])) > NNTPMAXGROUPLIST)
			tprintf ("Group list too long!  Group '%s' ignored!\n", argv[i]);
		else
			{
			if (i > 1)
				strcat (grouplist, ",");

			strcat (grouplist, argv[i]);
			}
		}

	Nntpgroups = strdup (grouplist);
	return 0;
	}

static int
    dobatch (argc, argv, p)
int argc;
char *argv[];
void *p;
	{
	if (argc < 2)
		{
		tprintf ("NNTP batch mode: %s, buffers: %d\n", nntpbatch ? "on" : "off", nntpbatchsize);
		return 0;
		}

	if (argc > 2)
		{
		nntpbatchsize = atoi (argv[2]);
		}

	return setbool (&nntpbatch, "NNTP batch mode", argc, argv);
	}

#ifdef	NNTP_RETRY	/*  GeK  */
static int
doretry (int argc, char *argv[], void *p)
{
	if (argc < 2)
	{
		tprintf ("NNTP retry when busy: %s, attempts: %u, delay: %us\n",
		         nntp_max_retries > 0 ? "on" : "off",
		         nntp_max_retries,
		         nntp_retry_delay);
		return (0);
	}
	
	if (!isdigit (argv[1][0]))
	{
		setbool ((int *) &nntp_max_retries, "NNTP retry mode", argc, argv);
		argc--;
		argv++;
	}

	if (argc > 1)
		nntp_max_retries = atoi (argv[1]);

	if (argc > 2)
		nntp_retry_delay = atoi (argv[2]);

	return (0);
}
#endif	/* NNTP_RETRY */

static int
    dosafety (argc, argv, p)
int argc;
char *argv[];
void *p;
	{
	return setbool (&nntpsafety, "NNTP safety mode", argc, argv);
	}

static int
    doverbose (argc, argv, p)
int argc;
char *argv[];
void *p;
	{
	return setbool (&nntpverbose, "NNTP verbose mode", argc, argv);
	}

static int
    donewgroups (argc, argv, p)
int argc;
char *argv[];
void *p;
	{
	return setbool (&nntpnewgroups, "NNTP get new groups", argc, argv);
	}

/* This is the routine that gets called every so often to connect to
 * NNTP servers.
 */

static void
     nntptick (tp)
void *tp;
	{
	/*  GeK - reduced stack from 3072 to 1536 *words* to reclaim some mem  */
	newproc ("NNTP client main", 1536, nntp_main, 0, tp, NULL, 0);
	}

struct textline
	{
	struct textline *nextline;	/* pointer to next text line */
	char text[NNTPMAXLEN];	 /* array holding text (shrunk to fit) */
	};

#define NULLTEXTLINE (struct textline *) 0
#define MINTEXTLINESIZE (sizeof(struct textline)-NNTPMAXLEN)

struct nntprequest
	{
	char msgid[MAX_MESSAGE_ID_LEN];	 /* id of requested header/body/article */	/*  GeK - save 256 bytes here  */
	struct textline *header; /* header text */
	unsigned int lines;		 /* lines in header/body/article */
	unsigned long chars;	 /* characters in header/body/article */
	enum nntp_request_status	/*  GeK - added enum tag to allow reuse later  */
		{
		reqempty = 0,		 /* vacant request */
		reqgetarticle,		 /* article from get file */
		reqnewarticle,		 /* article from new file */
		reqheader,			 /* header only */
		reqbody,			 /* body to accompany header */
		reqkill,			 /* kill this header */
		reqkeep,			 /* get a body to accompany this header */
		reqend,				 /* no more requests */
		reqquit				 /* close connection */
		}    status;
	char *newsgroup_list;
	struct nntprequest *nextrequest;	/* pointer to next request */
	};

#define NULLNNTPREQUEST (struct nntprequest *) 0

struct nntpsession
	{
	int s;					 /* NNTP socket */

	struct proc *childproc;	 /* child process, either init or send */

	FILE *getmsgf;			 /* file holding messages to get */
	FILE *newmsgf;			 /* temporary file for NEWNEWS response */
	FILE *historyf;			 /* history file */
	unsigned char *hashtable;/* history hash table */

	struct nntprequest
	           *requesthead, /* request pointer used by sender */
	           *requesttail; /* request pointer used by receiver */

	struct textline
	        *killlist[27],		 /* kill patterns */
	        *keeplist[27];		 /* keep patterns */

	enum nntp_request_status	init_article_status;	/* GeK - initial state for each article
														 *   reqnewarticle - default, no kill file 
														 *   reqheader     - kill file, header only
														 *   reqkill       - '*' entry in kill file
														 */

	int quit;				 /* if true quit immediately */

	unsigned int historyentries;		/* number of entries in history file */
	unsigned int historyscans;	/* number of complete history file scans */
	unsigned int getarticles;/* number of lines in get file */
	unsigned int getinvalid; /* number of invalid ids from get file */
	unsigned int getunavailable;		/* number of articles from get file
										 * unavailable */
	unsigned int getreceived;/* number of complete articles received */
	unsigned int newarticles;/* number of new article message ids received */
	unsigned int newduplicates;	/* number of duplicate articles */
	unsigned int newunavailable;		/* number of articles unavailalable */
	unsigned int newheadersreceived;	/* number of headers only received */
	unsigned int newreceived;/* number of complete articles received */
	unsigned int groupsreceived;		/* number of new groups received */
	unsigned long receivedarticlechars;	/* number of characters in articles
										 * received */
	unsigned long receivednewschars;	/* number of characters in news
										 * altogether */
	unsigned long receivedtotalchars;	/* number of characters received
										 * altogether */
	unsigned long startsession;	/* clock value at start of session */
	unsigned long startarticles;		/* clock value at start of article
										 * fetch */
	unsigned long endsession;/* clock value at end of session */
	};

#define NULLNNTPSESSION (struct nntpsession *) 0

static struct nntpsession *nntp = NULLNNTPSESSION;

static void
     error (level, text)
int level;
char *text;
	{
	char *p;							/* -> text							*/
	int	len;

	/*  GeK - only malloc when needed here ...  */
	p = text;
	len = strlen (text);
	if (len && text[len-1] == '\n')
	{
		p = mallocw (len);
		strncpy (p, text, len-1);
		p[len-1] = '\0';
	}
	if (level <= 2)
		/*
		 *  GeK - from Niel Kempson
		 *  Avoids NULL nntp ptr when outside time window
		 */
		log (((nntp == NULL) ? -1 : nntp->s), "%s", p);

	if (level <= nntptrace || nntpverbose != 0)
		tprintf ("%s\n", p);

	if (p != text)
		free (p);
	}

static void
     freetextlinelist (list)
struct textline **list;
	{
	struct textline *lastline, *thisline = *list;

	while (thisline != NULLTEXTLINE)
		{
		lastline = thisline;
		thisline = thisline->nextline;
		free (lastline);
		}

	*list = NULLTEXTLINE;
	}

static void
     createkilllists ()
	{
	FILE *killf;
	char buf[NNTPMAXLEN], *s, *d;
	int keep, n, keeplines = 0, killlines = 0;
	struct textline *newline, *lastkillline[27], *lastkeepline[27];
	int	index;	/*  GeK  */

	/*  GeK - init arrays to NULL  */
	memset (lastkillline, 0, 27 * sizeof (struct textline *));
	memset (lastkeepline, 0, 27 * sizeof (struct textline *));

	/*  GeK - set default article request status to get complete article  */
	nntp->init_article_status = reqnewarticle;

	/* If it exists, open the kill file */
	sprintf (buf, "%s/kill", Newsdir);
	if (access (buf, 0) == 0)
		{

		if (mlock (Newsdir, "kill"))
			{
			sprintf (buf, "NNTP Can't lock file %s/kill", Newsdir);
			error (2, buf);
			nntp->quit = TRUE;
			return;
			}

		if ((killf = fopen (buf, READ_TEXT)) == NULLFILE)
			{
			sprintf (buf, "NNTP Can't open file %s/kill", Newsdir);
			error (1, buf);
			rmlock (Newsdir, "kill");
			nntp->quit = TRUE;
			return;
			}

		/*  GeK - kill file exists, default to getting headers  */
		nntp->init_article_status = reqheader;

		while (fgets (buf, NNTPMAXLEN, killf) != NULLCHAR)
			{
			/* strip out extra characters and force lower case */

			s = d = buf;
			n = 0;
			keep = (*s == '!');
			if (keep)
				s++;

			while (*s != '\n' && *s != '\0')
				{
				if (isspace (*s))
					{
					do
						s++;
					while (isspace (*s));

					*d++ = ' ';
					}
				else
					{
					*d++ = tolower (*s++);
					}

				n++;
				}

			*d = '\0';

			/*  Skip comments and blank lines  */
			if (buf[0] == '#' || buf[0] == '\0' || (buf[0] == ' ' && buf[1] == '\0'))
				{
				if (!nntp->quit && !main_exit)
					pwait (NULL);
				continue;
				}

			if (buf[0] == ' ')
				index = 26;
			else
				index = tolower (buf[0]) - 'a';

			if (index < 0 || index > 26)
				{
				/*  GeK - detect '*'  */
				if (!keep && buf[0] == '*' && buf[1] == '\0')
					/*  Don't bother adding to kill list, just change initial status to kill  */
					nntp->init_article_status = reqkill;
				else
					/*  buf should already be 'ripped' here  */
					tprintf ("NNTP Kill: ignoring '%s%s'\n", (keep ? "!" : ""), buf);

				if (!nntp->quit && !main_exit)
					pwait (NULL);
				continue;
				}

			if ((newline = (struct textline *) malloc (MINTEXTLINESIZE + n + 1)) == NULLTEXTLINE)
				{
				error (2, "NNTP Can't allocate kill buffer");
				fclose (killf);
				rmlock (Newsdir, "kill");
				nntp->quit = TRUE;
				return;
				}

			newline->nextline = NULLTEXTLINE;
			memcpy (newline->text, buf, n + 1);

			if (keep)
				{
				if (lastkeepline[index] == NULLTEXTLINE)
					nntp->keeplist[index] = newline;
				else
					lastkeepline[index]->nextline = newline;

				lastkeepline[index] = newline;
				keeplines++;
				}
			else
				{
				if (lastkillline[index] == NULLTEXTLINE)
					nntp->killlist[index] = newline;
				else
					lastkillline[index]->nextline = newline;

				lastkillline[index] = newline;
				killlines++;
				}

			if (!nntp->quit && !main_exit)
				pwait (NULL);

			}
		if (nntptrace >= 3)
			tprintf ("NNTP Kill file read: %d kill lines, %d keep lines\n", killlines, keeplines);

		fclose (killf);
		rmlock (Newsdir, "kill");
		}
	}

static void
     scankilllists (line, status)
char *line;
int *status;
	{
	char buf[NNTPMAXLEN], *s, *d;
	struct textline *thisline;
	int	index;					/*  GeK - index of kill rule  */
	int	seen_colon = FALSE;		/*  GeK - for checking broken header lines  */

	/*
	 *  GeK - try and eliminate line as quickly as possible here
	 *  Trying to optimize for speed - common case of upper case letter first
	 */
	index = line[0] - 'A';
	if (index > 25 || index < 0)
		{
		if (index == (' ' - 'A'))
			index = 26;
		else
			index += 'A' - 'a';		/*  Adjust for lower case  */

		if (index < 0 || index > 26)
			/*  Can't match our kill lists so return now  */
			return;
		}

	if (nntp->keeplist[index] == NULL && (nntp->killlist[index] == NULL || *status == reqkill))
		return;	/*  No chance of changing things  */

	/* strip space and convert to lower case */
	s = line;
	d = buf;
	while (*s != '\n' && *s != '\0')
		{
		if (isspace (*s))
			{
			do
				s++;
			while (isspace (*s));

			*d++ = ' ';
			}
		else
			{
			if (*s == ':')
				seen_colon = TRUE;	/*  GeK - check for header line  */
			*d++ = tolower (*s++);
			}
		}

	*d = '\0';

	/*  GeK - long header line, no hope of matching accurately  */
	if (seen_colon == FALSE && index != 26)
		return;

	thisline = nntp->keeplist[index];
	while (thisline != NULLTEXTLINE)
		{
		if (wildmat (buf, thisline->text, NULLCHARP))
			{
			if (nntptrace >= 3)
				tprintf ("Keep: %s\n      %s\n", buf, thisline->text);

			*status = reqkeep;
			return;
			}

		thisline = thisline->nextline;
		}

	if (*status != reqkill)
		{
		thisline = nntp->killlist[index];
		while (thisline != NULLTEXTLINE)
			{
			if (wildmat (buf, thisline->text, NULLCHARP))
				{
				if (nntptrace >= 3)
					tprintf ("Kill: %s\n      %s\n", buf, thisline->text);

				*status = reqkill;
				return;
				}

			thisline = thisline->nextline;
			}
		}

	}

#define HASH1 32749
#define HASH2 32719

unsigned int
    hashmsgid (msgid, modulus)
char *msgid;
unsigned int modulus;
	{
	unsigned long int i = 0;

	while (*msgid != '\0')
		{
		do
			i = (i << 8) | *msgid++;
		while (*msgid != '\0' && i < (1L << 24));

		i %= modulus;
		}

	return (unsigned int) i;
	}

static void
     createhashtable ()
	{
	char msgid[NNTPMAXLEN];
	unsigned char *hashtable = NULL;
	unsigned int i, lines = 0, collisions = 0;
	int found;

	/* If there is sufficient memory then build a 64K bit hash table from
	 * the history file. */

	if (availmem () > Memthresh + (8192 * sizeof (unsigned char)) &&
		(hashtable = (unsigned char *) calloc (8192, sizeof (unsigned char))) != NULL)
		{
		rewind (nntp->historyf);
		while (fgets (msgid, NNTPMAXLEN, nntp->historyf) != NULL)
			{
			lines++;
			found = 1;
			i = hashmsgid (msgid, HASH1);
			if (!(hashtable[i / 4] & (0x01 << (i % 4))))
				{
				hashtable[i / 4] |= 0x01 << (i % 4);
				found = 0;
				}

			i = hashmsgid (msgid, HASH2);
			if (!(hashtable[i / 4] & (0x10 << (i % 4))))
				{
				hashtable[i / 4] |= 0x10 << (i % 4);
				found = 0;
				}

			if (found)
				collisions++;

			if (!nntp->quit && !main_exit)
				pwait (NULL);

			}
		rewind (nntp->historyf);
		if (nntptrace >= 3)
			tprintf ("NNTP Hash table created: %u ids, %u collisions\n", lines, collisions);

		}

	nntp->hashtable = hashtable;
	nntp->historyentries = lines;
	nntp->historyscans = 1;
	}

static int
    isinhashtable (msgid)
char *msgid;
	{
	unsigned int i, j;

	if (nntp->hashtable == NULL)
		return 1;
	else
		{
		i = hashmsgid (msgid, HASH1);
		j = hashmsgid (msgid, HASH2);
		if ((nntp->hashtable[i / 4] & (0x01 << (i % 4))) &&
			(nntp->hashtable[j / 4] & (0x10 << (j % 4))))
			return 1;

		}
	return 0;
	}

static int
    isinhistory (msgid)
char *msgid;
	{
	char historyid[NNTPMAXLEN];
	long historyfpos;
	int	count = 5;

	if (isinhashtable (msgid))
		{
		historyfpos = ftell (nntp->historyf);
		while (fgets (historyid, NNTPMAXLEN, nntp->historyf) != NULLCHAR)
			{
			if (FAST_STREQ (historyid, msgid))		/*  GeK  */
				return 1;

			if (--count == 0 && !nntp->quit && !main_exit)
				{
				pwait (NULL);
				count = 5;
				}

			}
		nntp->historyscans++;
		rewind (nntp->historyf);
		/*  GeK - moved ftell check to start of loop to prevent extra scans if fpos == 0  */
		while (ftell (nntp->historyf) < historyfpos
		       && fgets (historyid, NNTPMAXLEN, nntp->historyf) != NULLCHAR)
			{
			if (FAST_STREQ (historyid, msgid))
				return 1;

#ifdef	notdef	/*  GeK - moved to while() check  */
			if (ftell (nntp->historyf) >= historyfpos)
				break;
#endif

			if (--count == 0 && !nntp->quit && !main_exit)
				{
				pwait (NULL);
				count = 5;
				}

			}
		}

	return 0;
	}

static void
     addtohashtable (msgid)
char *msgid;
	{
	unsigned int i, j;

	if (nntp->hashtable != NULL)
		{
		i = hashmsgid (msgid, HASH1);
		nntp->hashtable[i / 4] |= 0x01 << (i % 4);
		j = hashmsgid (msgid, HASH2);
		nntp->hashtable[j / 4] |= 0x10 << (j % 4);
		}

	}

static void
     addtohistory (msgid)
char *msgid;
	{
	long historyfpos;
	int duphandle;

	addtohashtable (msgid);

	historyfpos = ftell (nntp->historyf);
	fseek (nntp->historyf, 0L, SEEK_END);
	fputs (msgid, nntp->historyf);
	if (nntpsafety)
		{
		/*  GeK - moved fflush() inside check - fseek() below will fflush() anyway  */
		fflush (nntp->historyf);
		duphandle = dup (fileno (nntp->historyf));
		close (duphandle);
		}

	fseek (nntp->historyf, historyfpos, SEEK_SET);
	nntp->historyentries++;
	}

/*
*	stripgroups reformats a group list in situ for a NEWNEWS request
*
*	returns -1 if there is leading space,
*				0 if line is blank otherwise
*				n the number of characters in the resultant string
*/

static int
    stripgroups (s)
char *s;
	{
	char *d = s;
	int count = 0;

	if (!isspace (*s))
		return -1;

	do
		s++;
	while (*s == ',' || isspace (*s));

	while (*s != '\0')
		{
		if (count != 0)
			{
			*d++ = ',';
			count++;
			}

		do
			{
			*d++ = *s++;
			count++;
			} while (*s != '\0' && *s != ',' && !isspace (*s));

		while (*s == ',' || isspace (*s))
			s++;
		}

	*d = '\0';
	return count;
	}

static unsigned long
     unitspersecond (units, milliseconds)
unsigned long units, milliseconds;
	{
	if (milliseconds == 0)
		{										 /* Avoid divide-by-zero */
		return 0;
		}
	else
		{
		if (units < 4294967UL)
			{
			return (units * 1000UL) / milliseconds;
			}
		else
			{									 /* Avoid overflow */
			return units / (milliseconds / 1000UL);
			}
		}
	}

/*
 *  GeK - Core that was in getreply() extracted for flexibility
 *  Returns response in specified buffer, doesn't log error responses (>= 400)
 *  Removes trailing \n from response buffer
 */
static int
getreply_in_buf (char *buf)
{
	int n, response;

	while ((n = recvline (nntp->s, buf, NNTPMAXLEN)) > 0)
	{
		nntp->receivedtotalchars += n + 1;	/*  GeK - \r  */
		if (nntptrace >= 3)
			tprintf ("<==%s", buf);

		/* skip informative messages and blank lines */
		if (buf[0] == '\n' || buf[0] == '1')
			continue;

		/*  Remove trailing \n  */
		if (buf[n-1] == '\n')
			buf[n-1] = '\0';

		response = atoi (buf);
		return (response);
	}
	return (-1);
}

/*  GeK - rewritten to use getreply_in_buf()  */
static int
getreply (void)
{
	char buf[NNTPMAXLEN];
	int  response;

	response = getreply_in_buf (buf);

	if (response >= 400)
	{
		/* Display error message. */
		error (2, buf);
	}

	return (response);
}

static int
    getheader (request)
struct nntprequest *request;
	{
	char buf[NNTPMAXLEN];
	int n, nlines = 0;
	unsigned long nchars = 0L;
	struct textline *newline, *lastheaderline = NULLTEXTLINE;
	unsigned long mstime = msclock ();	/*  GeK  */
	int last_had_eol = TRUE;	/*  GeK - start of new line, last line had \n  */

	while ((n = recvline (nntp->s, buf, NNTPMAXLEN)) != -1)
		{
		/*  GeK - account for \r  */
		nntp->receivedtotalchars += n + ((buf[n-1] == '\n') ? 1 : 0);
		nchars += n + ((buf[n-1] == '\n') ? 1 : 0);

		/*  GeK - workaround for \r\n sequence written to batch.txt breaking cppnews  */
		if (n > 1 && buf[n-2] == '\r')
			{
			n--;
			buf[n-1] = buf[n];
			buf[n] = '\0';
			}

		if (nntptrace >= 4)
			tprintf ("<==%s", buf);

		if (last_had_eol && (buf[0] == 'N' || buf[0] == 'n')
		    && strnicmp (buf+1, "ewsgroups:", 10) == 0
		    && request->newsgroup_list == NULL)
			{
			request->newsgroup_list = strdup (buf+12);
			if (buf[n-1] == '\n' && n > 12)	/*  Trim off any \n  */
				request->newsgroup_list[n-13] = '\0';
			}
		else
		if (last_had_eol && buf[0] == '.')
		  if (buf[1] == '\n')
			{
			request->lines += nlines;
			if (nntptrace >= 3)
				{
				mstime = msclock () - mstime;
				/*  GeK - display ms rather than secs here (sec is usually 0)  */
				tprintf ("<--%d lines, %lu chars, %lu ms (%lu bytes/sec)\n",
						 nlines,
						 nchars,
						 mstime,
						 unitspersecond (nchars, mstime));
				}
			/* default to getting the body */

			if (request->status != reqkill)
				request->status = reqkeep;

			return nlines;
			}
		  else		/*  escaped '.' characters  */
			{
			memmove (buf, buf + 1, n);
			n--;
			}

		last_had_eol = (buf[n-1] == '\n');

		if (last_had_eol)	/*  GeK - only count complete lines  */
			nlines++;

		request->chars += n;

		if ((newline = (struct textline *) malloc (MINTEXTLINESIZE + n + 1)) == NULLTEXTLINE)
			{
			error (2, "NNTP Can't allocate header buffer");
			return -1;
			}

		newline->nextline = NULLTEXTLINE;
		memcpy (newline->text, buf, n + 1);

		if (lastheaderline == NULLTEXTLINE)
			request->header = newline;
		else
			lastheaderline->nextline = newline;

		lastheaderline = newline;

		if (request->status != reqkeep)
			scankilllists (buf, (int *) &(request->status));

		}
	sprintf (buf, "NNTP Receive error after %d lines", nlines);
	error (1, buf);
	return -1;
	}

static void
     moveheadertofile (fp, request)
FILE *fp;
struct nntprequest *request;
	{
	struct textline *doneline, *thisline = request->header;
	int	count = 5;

	while (thisline != NULLTEXTLINE)
		{
		fputs (thisline->text, fp);
		doneline = thisline;
		thisline = thisline->nextline;
		free (doneline);
		/*  GeK - added to prevent slow write affecting other processes  */
		if (--count == 0 && !nntp->quit && !main_exit)
			{
			pwait (NULL);
			count = 5;
			}
		}

	request->header = NULLTEXTLINE;
	}

static int
    gettxt (fp, request, newnews)
FILE *fp;
struct nntprequest *request;
int newnews;
	{
	char buf[NNTPMAXLEN];
	int n, nlines = 0;
	unsigned long nchars = 0L;
	unsigned long mstime = msclock ();	/*  GeK  */
	int	last_had_eol = TRUE;	/*  GeK  */

	while ((n = recvline (nntp->s, buf, NNTPMAXLEN)) != -1)
		{
		/*  GeK - account for \r  */
		nntp->receivedtotalchars += n + ((buf[n-1] == '\n') ? 1 : 0);
		nchars += n + ((buf[n-1] == '\n') ? 1 : 0);

		/*  GeK - workaround for \r\n sequence written to batch.txt breaking cppnews  */
		if (n > 1 && buf[n-2] == '\r')
			{
			n--;
			buf[n-1] = buf[n];
			buf[n] = '\0';
			}

		if (nntptrace >= 4)
			tprintf ("<==%s", buf);

		if (last_had_eol && (buf[0] == 'N' || buf[0] == 'n')
		    && strnicmp (buf+1, "ewsgroups:", 10) == 0
		    && !newnews
		    && request->newsgroup_list == NULL)
			{
			request->newsgroup_list = strdup (buf+12);
			if (buf[n-1] == '\n' && n > 12)	/*  Trim off any \n  */
				request->newsgroup_list[n-13] = '\0';
			}
		else
		if (last_had_eol && buf[0] == '.')
		  if (buf[1] == '\n')
			{
			if (nntptrace < 4 && (newnews != 0 && nntpverbose != 0))
				{
				/*  GeK - show last one  */
				if (nlines > 0)
					tprintf ("\rNews message-id %d received ", nlines);
				tputc ('\n');
				}
				
			if (request != NULLNNTPREQUEST)
				request->lines += nlines;

			if (nntptrace >= 3)
				{
				mstime = msclock () - mstime;	/*  GeK  */
				/*  GeK - ms rather than secs  */
				tprintf ("<--%d lines, %lu chars, %lu ms (%lu bytes/sec)\n",
						 nlines,
						 nchars,
						 mstime,
						 unitspersecond (nchars, mstime));
				}

			return nlines;
			}
		  else		/*  escaped '.' characters  */
			{
			memmove (buf, buf + 1, n);
			n--;
			}

		last_had_eol = (buf[n-1] == '\n');

		if (last_had_eol)	/*  GeK - only count complete lines  */
			nlines++;

		/*  GeK - don't display every id for speed  */
		if (nntptrace < 4 && (newnews != 0 && nntpverbose != 0)
		    && (nlines < 20
		        || (nlines < 200 && !(nlines % 20))
		        || !(nlines % 50)))
			{
			tprintf ("\rNews message-id %d received ", nlines);
			}

		if (request != NULLNNTPREQUEST)
			request->chars += n;

		/*  GeK - throw away message-ids too long  */
		if (newnews && n > (MAX_MESSAGE_ID_LEN - 1))
			{
			char	*p;

			p = malloc (n+32);	/*  Don't mallocw in case of low mem  */
			if (p != NULL)
				{
				if (last_had_eol)
					buf[n-1] = '\0';	/*  rip (buf);  */
				sprintf (p, "NNTP ID too long: %s", buf);
				error (2, p);
				free (p);
				}

			/*  stats need adjusting here  */
			nntp->newunavailable++;

			continue;	/*  while (recvline)  */
			}

		/*  GeK - check for disk full  */
		if (fputs (buf, fp) == EOF)
			break;	/*  while (recvline)  */
		}

	if (nntptrace < 4 && (newnews != 0 && nntpverbose != 0))
		{
		tputc ('\n');
		}

	sprintf (buf, "NNTP Receive error after %d lines", nlines);
	error (1, buf);
	return -1;
	}

static int
    putarticle (msgf, msgfpos, request)
FILE *msgf;
long int *msgfpos;
struct nntprequest *request;

#if NEWSBATCH				 /* articles saved for batching news reader */
	{
#ifdef	notdef	/*  GeK - simplified  */
	char buf[NNTPMAXLEN];
#endif	/* notdef */
	int duphandle;

	(void) fflush (msgf);

	if (ferror (msgf))
		{
		error (1, "NNTP Error writing newsbatch file");
		return -1;
		}

	(void) fseek (msgf, *msgfpos, SEEK_SET);
	fprintf (msgf, NEWSBATCHFORMAT, request->chars);
	/*  GeK - next fseek() will fflush() anyway, moved fflush inside safety check  */

	if (nntpsafety)
		{
		fflush (msgf);
		duphandle = dup (fileno (msgf));
		close (duphandle);
		}

	if (nntpverbose)
		{
#ifdef	notdef	/*  GeK - simplified  */
		int	last_had_eol = TRUE;
		char *p = buf;
		int line_count = 5;

		(void) fseek (msgf, *msgfpos, SEEK_SET);
		while (fgets (buf, NNTPMAXLEN, msgf) != NULLCHAR)
			{
			p = strchr (buf, '\n');
			if (last_had_eol &&
			    (buf[0] == '\n' || strnicmp (buf, "Newsgroups:", 11) == 0))
				break;
			last_had_eol = (p != NULL);
			/*  GeK - give up some time now and again  */
			if (--line_count == 0)
				{
				if (!nntp->quit && !main_exit)
					pwait (NULL);
				line_count = 5;
				}
			}

		if (strnicmp (buf, "Newsgroups:", 11) == 0)
		    {
		    if (p != NULL)
				*p = '\0';
		    }
		else
			buf[12] = '\0';
#endif	/* notdef */

		tprintf ("News arrived (%u/%u): %s, %s",
		nntp->getreceived + nntp->newheadersreceived + nntp->newreceived + 1,
			 (nntp->getarticles - nntp->getinvalid - nntp->getunavailable) +
		   (nntp->newarticles - nntp->newduplicates - nntp->newunavailable),
#ifdef	notdef	/*  GeK - simplified  */
				 buf + 12,
#endif
				 request->newsgroup_list,
				 request->msgid);
		}

	/*  GeK  */
	free (request->newsgroup_list);
	request->newsgroup_list = NULL;

	/* set new length */

	(void) fseek (msgf, 0L, SEEK_END);
	*msgfpos = ftell (msgf);
	fprintf (msgf, NEWSBATCHFORMAT, 0UL);

	return 0;
	}

#else						 /* articles saved for mail style news reader */
	{
	char buf[NNTPMAXLEN], froml[NNTPMAXLEN], newgl[NNTPMAXLEN];
	FILE *fp;
	char *cp;
	int lines;

	/* convert the article into mail format */
	rewind (msgf);
	froml[0] = '\0';
	newgl[0] = '\0';
	while (fgets (buf, NNTPMAXLEN, msgf) != NULLCHAR)
		{
		if (strnicmp (buf, "From: ", 6) == 0)
			{
			time_t t;

			rip (&buf[6]);
			time (&t);
			sprintf (froml, "From %s %s", &buf[6], ctime (&t));
			if (newgl[0] != '\0')
				break;

			}

		if (strnicmp (buf, "Newsgroups: ", 12) == 0)
			{
			strcpy (newgl, &buf[12]);
			rip (newgl);
			if (froml[0] != '\0')
				break;

			}
		/* invalid article - missing 'From:' or 'Newsgroups:' line */

		if (strcmp (buf, "\n") == 0 && (froml[0] == '\0' || newgl[0] == '\0'))
			{
			if (nntptrace >= 2)
				tprintf ("NNTP Invalid article received: %s, article %s", newgl, request->msgid);

			return 0;
			}

		}
	sprintf (buf, "%s/", News_spool ? News_spool : Mailspool);
	for (cp = newgl;; ++cp)
		{
		if (*cp == '.')
			{
#ifdef UNIX
			mkdir (buf, 0755);					 /* create a subdirectory, if
												  * necessary */
#else
			mkdir (buf);						 /* create a subdirectory, if
												  * necessary */
#endif
			strcat (buf, "/");
			continue;
			}

		if (*cp == ',' || *cp == '\0')
			{
			char tempdir[80], prefix[20], *p;

			strcpy (tempdir, buf);
			if ((p = strrchr (tempdir, '/')) != NULLCHAR)
				{
				*p++ = '\0';
				strcpy (prefix, p);
				}

			if (mlock (tempdir, prefix))
				{
				if (nntptrace >= 2)
					tprintf ("NNTP Group '%s' is locked\n", buf);

				return -1;
				}

			strcat (buf, ".txt");
			/* open the mail file */
			if (nntptrace >= 2)
				tprintf ("Writing article to '%s'\n", buf);

			if ((fp = fopen (buf, APPEND_TEXT)) != NULLFILE)
				{
				fputs (froml, fp);
				rewind (msgf);
				lines = request->lines;
				while ((fgets (buf, NNTPMAXLEN, msgf) != NULLCHAR) && lines--)
					{
					/* for UNIX mail compatiblity */
					if (strncmp (buf, "From ", 5) == 0)
						putc ('>', fp);

					fputs (buf, fp);
					}

				putc ('\n', fp);
				fclose (fp);
				}

			rmlock (tempdir, prefix);
			if (*cp == '\0')
				break;
			else
				sprintf (buf, "%s/", News_spool ? News_spool : Mailspool);

			continue;
			}

		buf[strlen (buf) + 1] = '\0';
		buf[strlen (buf)] = strchr (validchars, tolower (*cp)) ? *cp : '_';
		}

	rewind (msgf);

	if (nntpverbose)
		tprintf ("News arrived (%u/%u): %s, %s",
		nntp->getreceived + nntp->newheadersreceived + nntp->newreceived + 1,
			 (nntp->getarticles - nntp->getinvalid - nntp->getunavailable) +
		   (nntp->newarticles - nntp->newduplicates - nntp->newunavailable),
				 buf + 12,
				 request->msgid);

	return 0;
	}

#endif

static void
     nntp_init (argc, argv, parentproc)
int argc;
void *argv, *parentproc;
	{
	createhashtable ();
	createkilllists ();

	psignal (parentproc, 0);
	nntp->childproc = NULLPROC;
	}

static void
     nntp_send (argc, argv, parentproc)
int argc;
void *argv, *parentproc;
	{
	int quit = FALSE;

	if (nntp->getmsgf != NULLFILE)
		rewind (nntp->getmsgf);

	rewind (nntp->newmsgf);

	while (!quit)
		{

		switch (nntp->requesthead->status)
			{

			case reqempty:
				if (nntp->getmsgf != NULLFILE)
					{
					while (fgets (nntp->requesthead->msgid, MAX_MESSAGE_ID_LEN, nntp->getmsgf) != NULLCHAR)
						{						 /* Find a line with a valid
												  * message id */
						char *cp = nntp->requesthead->msgid;

						nntp->getarticles++;
						if (*cp++ != '<')
							{
							nntp->getinvalid++;
							if (!nntp->quit && !main_exit)
								pwait (NULL);

							continue;
							}

						while (*cp != '>' && isgraph (*cp))
							cp++;

						if (*cp++ == '>')
							{
							*cp++ = '\n';
							*cp = '\0';
							nntp->requesthead->status = reqgetarticle;
							break;
							}

						nntp->getinvalid++;
						if (!nntp->quit && !main_exit)
							pwait (NULL);
						else
							break;		/*  GeK - exit get loop if terminating  */
						}
					}
				if (nntp->requesthead->status == reqempty)
					{
					while (fgets (nntp->requesthead->msgid, MAX_MESSAGE_ID_LEN, nntp->newmsgf) != NULLCHAR)
						{
						struct nntprequest *request = nntp->requesthead->nextrequest;
						int                 found   = 0;

						/*
						 *  GeK - Demon's NewsBorg implementation does not de-dupe newnews ids for
						 *  cross-posted articles;  if a request for a dupe is outstanding in the
						 *  ring buffer then it won't be caught so search the ring buffer.  Ick.
						 *  The same thing could happen across multiple NEWNEWS requests anyway.
						 *  Assume it's quicker to search ring buffer first, then history check.
						 */
						while (request != nntp->requesthead)
							{	
							if (request->status  /* > reqempty (==0) */
							    && request->status < reqend
							    && FAST_STREQ (request->msgid, nntp->requesthead->msgid))
								{
								if (nntpverbose)
									tputs ("New");	/*  Preceeds dupe message below  */

								found = 1;
								break;	/*  while (request list)  */
								}
							request = request->nextrequest;
							}

						if (!found && !isinhistory (nntp->requesthead->msgid))
							{
#ifdef	notdef	/*  GeK  */
 							if (nntp->killlist == NULLTEXTLINE)
 								nntp->requesthead->status = reqnewarticle;
 							else
 								nntp->requesthead->status = reqheader;
#endif
							/*  GeK - prime status from initial state  */
							nntp->requesthead->status = nntp->init_article_status;

							break;
							}

						if (nntpverbose)
							tprintf ("News duplicate: article %s", nntp->requesthead->msgid);

						nntp->newduplicates++;
						if (!nntp->quit && !main_exit)
							pwait (NULL);
						else
							break;		/*  GeK - exit duplicate loop if terminating  */
						}
					}

				if (nntp->requesthead->status == reqempty)
					nntp->requesthead->status = reqend;

				break;

			case reqkeep:
				nntp->requesthead->status = reqbody;
				break;

			case reqend:
				nntp->requesthead->status = reqquit;
				break;

			default:							 /* should never get here */
				error (2, "NNTP Send internal error 1");
				nntp->requesthead->status = reqquit;
			}

#ifdef	notdef	/*  GeK - moved below exit check  */
		if (!nntpbatch && nntp->requesthead != nntp->requesttail)
			pwait (Curproc);
#endif

		if (nntp->quit || main_exit)
			{
			psignal (parentproc, 0);
			break;
			}

		/*
		 *  GeK - moved this from above terminating check above
		 *      - will pwait() return here if main_exit is true ?
		 *      - should the return value of pwait() be checked ?
		 */
		if (!nntpbatch && nntp->requesthead != nntp->requesttail)
			pwait (Curproc);

		switch (nntp->requesthead->status)
			{
			case reqend:
				break;
			case reqquit:
				if (nntptrace >= 3)
					tprintf ("==>QUIT\n");

				usprintf (nntp->s, "QUIT\n");
				quit = TRUE;
				break;
			case reqgetarticle:
			case reqnewarticle:
				if (nntptrace >= 3)
					tprintf ("==>ARTICLE %s", nntp->requesthead->msgid);

				usprintf (nntp->s, "ARTICLE %s", nntp->requesthead->msgid);
				break;
			case reqheader:
				if (nntptrace >= 3)
					tprintf ("==>HEAD %s", nntp->requesthead->msgid);

				usprintf (nntp->s, "HEAD %s", nntp->requesthead->msgid);
				break;
			case reqbody:
				if (nntptrace >= 3)
					tprintf ("==>BODY %s", nntp->requesthead->msgid);

				usprintf (nntp->s, "BODY %s", nntp->requesthead->msgid);
				break;
			default:							 /* should never get here */
				error (2, "NNTP Send internal error 2");
				if (nntptrace >= 3)
					tprintf ("==>QUIT\n");

				usprintf (nntp->s, "QUIT\n");
				quit = TRUE;
			}

		if (nntpbatch && nntp->requesthead->nextrequest == nntp->requesttail)
			pwait (Curproc);

		nntp->requesthead = nntp->requesthead->nextrequest;
		psignal (parentproc, 0);

		}

	/* let the folks know we've gone */
	nntp->childproc = NULLPROC;
	}

static void
     nntp_main (i1, tp, v1)
int i1;
void *tp, *v1;
	{
	struct nntpservers *np = (struct nntpservers *) tp;
	struct sockaddr_in fsocket;
	int reply, lines, i;
	char buf1[NNTPMAXLEN], buf2[NNTPMAXLEN], *cp;
	FILE *tmpf1, *tmpf2, *gettmpf, *msgf = NULLFILE;
	long int tmpfpos, msgfpos;
	char datenow[14], lastdate[14];
#ifdef	NNTP_POST
	int nntp_posting_allowed = 0;	/*  GeK - from opening banner response  */
#endif
#ifdef	NNTP_RETRY
	unsigned int	retry_attempt = 0;
#endif
#ifdef	NNTP_SESSION
	struct session	*sp;
#endif

	if (nntptrace >= 3)
		tprintf ("NNTP daemon entered, target = %s\n", np->name);

	if (nntp != NULLNNTPSESSION)
		{										 /* nntp already active */
		error (2, "NNTP Already active");
		start_timer (&np->nntpcli_t);
		return;
		}

		{										 /* check the time - keep a
												  * record for updating the
												  * data file */
		time_t t;
		struct tm *ltm;
		int now;

		time (&t);								 /* more portable than
												  * gettime() */
		ltm = localtime (&t);
		now = ltm->tm_hour * 100 + ltm->tm_min;
		if ((np->lowtime < np->hightime) ?		 /* doesn't cross midnight */
			(now < np->lowtime || now >= np->hightime) :
			(now < np->lowtime && now >= np->hightime))
			{
			sprintf (buf2, "NNTP Window to '%s' not open", np->name);
			error (2, buf2);
			start_timer (&np->nntpcli_t);
			return;
			}
		t -= NEWNEWS_FUZZ;		/*  GeK - fudge factor now applied after checking NNTP window  */

		ltm = gmtime (&t);
		sprintf (datenow, "%02d%02d%02d %02d%02d%02d", ltm->tm_year % 100, ltm->tm_mon + 1,
				 ltm->tm_mday, ltm->tm_hour, ltm->tm_min, ltm->tm_sec);
		}

	if (availmem () < Memthresh + sizeof (struct nntpsession) ||
		(nntp = (struct nntpsession *) calloc (1, sizeof (struct nntpsession))) == NULLNNTPSESSION)
		{
		error (2, "NNTP Memory too low");
		start_timer (&np->nntpcli_t);
		return;
		}

#ifdef	NNTP_SESSION
	if ((sp = newsession (np->name, TELNET)) == NULL)
		{
		error (2, "Unable to start new NNTP session\n");
		return;
		}
	else
		{
		struct usock *up;

		up = itop (Curproc->output);
		if (up != NULL)
			up->noblock = 1;	/*  Prevent screen output from blocking  */

		/*  Transfer focus back to command session  */
		Lastcurr = Current;
		Current = Command;
		swapscreen (sp, Command);
		psignal (Command, 0);
		}
#endif	/* NNTP_SESSION */

	nntp->s = -1;

	if (nntpverbose)
		tprintf ("Trying to connect to news server %s ....\n", np->name);

	/* set up a circular list of request buffers */

		{
		struct nntprequest *head, *tail;

		if (availmem () < Memthresh + sizeof (struct nntprequest) ||
			(head = tail =
			 (struct nntprequest *) calloc (1, sizeof (struct nntprequest))) == NULLNNTPREQUEST)
			{
			error (2, "NNTP Memory too low");
			goto quit;
			}

		for (i = 1; i < nntpbatchsize || i < 2; i++)
			{
			if (availmem () < Memthresh + sizeof (struct nntprequest) ||
				(tail->nextrequest =
				 (struct nntprequest *) calloc (1, sizeof (struct nntprequest))) == NULLNNTPREQUEST)
				{
				error (2, "NNTP Memory too low");
				goto quit;
				}

			tail = tail->nextrequest;
			}

		nntp->requesttail = nntp->requesthead = tail->nextrequest = head;
		}

	/*  GeK - moved history and init from after successful connect  */

	/* Open the history file */

	sprintf (buf1, "%s/history", Newsdir);
	if (mlock (Newsdir, "history"))
		{
		sprintf (buf2, "NNTP Can't lock file %s", buf1);
		error (2, buf2);
		goto quit;
		}

	if ((nntp->historyf = fopen (buf1, APPEND_TEXT)) == NULLFILE)
		{
		sprintf (buf2, "NNTP Can't open file %s", buf1);
		error (1, buf2);
		rmlock (Newsdir, "history");
		goto quit;
		}

	/* Build internal tables */

	nntp->childproc = newproc ("NNTP client init", 1024, nntp_init, 0, NULL, Curproc, 0);

#ifdef	NNTP_RETRY
nntp_busy_loop:
	retry_attempt++;
#endif

	/* Find the server and try to connect */
	fsocket.sin_addr.s_addr = resolve (np->name);
	if (fsocket.sin_addr.s_addr == 0)
		{										 /* No IP address found */
		sprintf (buf2, "NNTP Can't resolve host: %s", np->name);
		error (2, buf2);
		goto quit;
		}

	fsocket.sin_family = AF_INET;
	fsocket.sin_port = IPPORT_NNTP;
	nntp->s = socket (AF_INET, SOCK_STREAM, 0);
#ifdef	NNTP_SESSION
	sp->s = nntp->s;
#endif
	/*  GeK - check for socket failing  */
	if (nntp->s < 0)
		{
		error (1, "NNTP socket failed");
		goto quit;
		}
	sockmode (nntp->s, SOCK_ASCII);

	alarm (connect_wait_val);
	if (connect (nntp->s, (char *) &fsocket, SOCKSIZE) == -1)
		{
		cp = (!run_timer (&Curproc->alarm)) ? "timeout" : sockerr (nntp->s);
		alarm (0L);
		sprintf (buf2, "NNTP Connect failed: %s", cp != NULL ? cp : "");
		error (2, buf2);
		goto quit;
		}

	alarm (0L);

#ifndef	NNTP_RETRY	/*  GeK - save connect msg until successful connect  */
	if (nntpverbose)
		tprintf ("Connected to news server\n");
#endif	/* !NNTP_RETRY */

	/*  GeK - history and init spawn was here  */

	/* Eat the banner */
	reply = getreply_in_buf (buf1);
	/* 
	 *  GeK - the old behaviour of getreply() ...
	 *	if (reply >= 400)
	 *		error (2, buf1);
	 *  .. is now fully logged below with the banner (after 20x check)
	 */

#ifdef	NNTP_RETRY

	if (reply == 400 && retry_attempt < nntp_max_retries)
		{
		/*
		 *  GeK - this fudge attempts to ensure that we've processed the FIN
		 *  from the server that should accompany the '400 busy'
		 *  - this way, we can almost guarantee that our TCB
		 *  won't end up the Time Wait state and eat up memory.
		 */
		alarm (500L);
		recvchar (nntp->s);
		alarm (0L);

		close_s (nntp->s);		/*  Close connection  */

		if (nntpverbose)
			tprintf ("NNTP server %s [%s] attempt %u retrying: %s\n", 
			         np->name, inet_ntoa (fsocket.sin_addr.s_addr),
			         retry_attempt, buf1);

		/*
		 *  There's the possibility here that KA9Q will exit (or crash)
		 *  during this pause, leaving the history locked
		 */
		if (nntp_retry_delay)
			pause ((long) nntp_retry_delay * 1000);

		goto nntp_busy_loop;
		}
	else
		{
		if (nntpverbose)
			tprintf ("NNTP Connected to %s [%s] on attempt %u\n",
			         np->name, inet_ntoa (fsocket.sin_addr.s_addr),
			         retry_attempt);
		}

	/*  Fall through to handle other responses */ 

#endif	/* NNTP_RETRY */

	if (!(reply == 200 || reply == 201))
		{
		if (reply < 0)
			error (1, "NNTP Connection unexpectedly closed");
		else
			{
			/*  GeK - log full banner  */
			sprintf (buf2, "NNTP Bad reply on banner: %s", buf1);
			error (1, buf2);
			}
		goto quit;
		}

#ifdef	NNTP_RETRY
	log (nntp->s, "NNTP open %s attempt %u: %s", np->name, retry_attempt, buf1);
#else
	log (nntp->s, "NNTP open %s: %s", np->name, buf1);
#endif

#ifdef	NNTP_POST
	nntp_posting_allowed = (reply != 201);
#endif

	/*  GeK - only start clock on successful connect  */
	nntp->startsession = msclock ();

#ifdef	NNTP_MODE_READER	/*  GeK - Mode Reader support  */
	cp = "==>MODE READER\n";
	if (nntptrace >= 3)
		tputs (cp);
	usputs (nntp->s, cp+3);
	/*  Ignore response  */
	getreply ();
#endif	/* NNTP_MODE_READER */

#ifdef	NNTP_AUTH	/*  GeK - Simple authentication AUTHINFO USER/PASS support  */
	if (np->username != NULL && np->password != NULL)
		{
		sprintf (buf1, "==>AUTHINFO USER %s\n", np->username);
		if (nntptrace >= 3)
			tputs (buf1);
		usputs (nntp->s, buf1 + 3);
		reply = getreply_in_buf (buf1);
		if (reply != 381 && reply != 350)	/*  381 or 350 - AUTH OK, PASS required  */
			{
			sprintf (buf2, "NNTP Auth username failed: %s", reply < 0 ? "Connection unexpectedly closed" : buf1);
			error (1, buf2);
			goto quit;
			}

		sprintf (buf1, "==>AUTHINFO PASS %s\n", np->password);
		if (nntptrace >= 3)
			tputs (buf1);
		usputs (nntp->s, buf1 + 3);
		reply = getreply_in_buf (buf1);
		if (reply != 281 && reply != 250)	/*  281 or 250 - PASS OK, authorized  */
			{
			sprintf (buf2, "NNTP Auth password failed: %s", reply < 0 ? "Connection unexpectedly closed" : buf1);
			error (1, buf2);
			goto quit;
			}

		if (nntpverbose)
			tprintf ("NNTP authentication accepted for %s\n", np->username);
		}
#endif	/* NNTP_AUTH */

#ifdef	NNTP_POST
	if (nntppost > 0)		/* SCT: Attempt NNTP posting */
		{
		if (!nntp_posting_allowed)
			{
			error (2, "NNTP Post: no permission");
			}
		/*  GeK - post outstanding articles  */
		else if (nntp_post () < 0)
			{
			error (1, "NNTP Post failed");
#ifdef notdef				/* SCT: don't give up before collecting! */
			goto quit;
#endif
			}
		/*  SCT - skip news download if "post only" mode (nntppost > 1)  */
		else if (nntppost > 1)
			goto quit;
		}
#endif	/* NNTP_POST */

	/* Open nntp.dat */

	sprintf (buf1, "%s/nntp.dat", Newsdir);
	if (mlock (Newsdir, "nntp"))
		{
		sprintf (buf2, "NNTP Can't lock file %s", buf1);
		error (2, buf2);
		goto quit;
		}

	if ((tmpf1 = fopen (buf1, APPEND_TEXT)) == NULLFILE)
		{
		sprintf (buf2, "NNTP Can't open file %s", buf1);
		error (1, buf2);
		rmlock (Newsdir, "nntp");
		goto quit;
		}

	/* Open temporary file for message ids */

	if ((nntp->newmsgf = tmpfile ()) == NULLFILE)
		{
		error (1, "NNTP Can't open temporary file");
		fclose (tmpf1);
		rmlock (Newsdir, "nntp");
		goto quit;
		}

	/* Look for a line in the data file with a matching server name */

	rewind (tmpf1);
	strcpy (lastdate, "700101 000000");
	i = strlen (np->name);
	cp = buf1 + i;
	while (fgets (buf1, NNTPMAXLEN, tmpf1) != NULLCHAR)
		{
		if (strnicmp (buf1, np->name, i) == 0 && isspace (*cp))
			{
			unsigned long date, time;

			if (sscanf (cp, " %6lu %6lu", &date, &time) == 2)
				sprintf (lastdate, "%06lu %06lu", date, time);

			break;
			}

		}

	/* Look for a newsgroup list in the data file */
	i = -1;
	while (fgets (buf2, NNTPMAXLEN, tmpf1) != NULLCHAR)
		if ((i = stripgroups (buf2)) != 0)
			break;

	/* if there is a newsgroup list in the data file */

	if (i > 0)
		{

		do
			{
			if (i > NNTPMAXGROUPLIST)
				{
				error (1, "NNTP Newsgroup line too long");
				fclose (tmpf1);
				rmlock (Newsdir, "nntp");
				goto quit;
				}

			strcpy (buf1, buf2);
			i = -1;

			/* add other entries until no more entries, blank line or the
			 * list is full */
			while ((fgets (buf2, NNTPMAXLEN, tmpf1) != NULLCHAR) && ((i = stripgroups (buf2)) > 0))
				if (strlen (buf1) + i >= NNTPMAXGROUPLIST)
					break;
				else
					{
					strcat (buf1, ",");
					strcat (buf1, buf2);
					i = -1;
					}

			/* send the request and get the response */
			if (nntptrace >= 3)
				tprintf ("==>NEWNEWS %s %s GMT\n", buf1, lastdate);

			usprintf (nntp->s, "NEWNEWS %s %s GMT\n", buf1, lastdate);

			if ((reply = getreply ()) != 230 || (lines = gettxt (nntp->newmsgf, NULL, 1)) == -1)
				{
				if (reply == -1 || lines == -1)
					error (1, "NNTP Connection closed while getting NEWNEWS");
				else
					{
					sprintf (buf2, "NNTP Error while getting NEWNEWS (response was %d)", reply);
					error (1, buf2);
					}
				fclose (tmpf1);
				rmlock (Newsdir, "nntp");
				goto quit;
				}

			nntp->newarticles += lines;

			/* if we're on a blank line then look for more newsgroup
			 * entries */
			if (i == 0)
				{
				i = -1;
				while (fgets (buf2, NNTPMAXLEN, tmpf1) != NULLCHAR)
					if ((i = stripgroups (buf2)) != 0)
						break;

				}

			} while (i > 0);

		}
	else
		{										 /* use the addservers entry,
												  * or the groups entry, or
												  * get everything (!) */

		/*  GeK - don't do a NEWNEWS *  */
		if (Nntpgroups == NULL)
			{
			error (1, "NNTP No newsgroups specified");
			fclose (tmpf1);
			rmlock (Newsdir, "nntp");
			goto quit;
			}

		if (nntptrace >= 3)
			tprintf ("==>NEWNEWS %s %s GMT\n",
					 np->groups ? np->groups : (Nntpgroups ? Nntpgroups : "*"), lastdate);

#ifdef	notdef	/*  GeK - removed this  */
		nntp_reg ();
#endif
		usprintf (nntp->s, "NEWNEWS %s %s GMT\n",
		np->groups ? np->groups : (Nntpgroups ? Nntpgroups : "*"), lastdate);

		if ((reply = getreply ()) != 230 || (lines = gettxt (nntp->newmsgf, NULL, 1)) == -1)
			{
			if (reply == -1 || lines == -1)
				error (1, "NNTP Connection closed while getting NEWNEWS");
			else
				{
				sprintf (buf2, "NNTP Error while getting NEWNEWS (response was %d)", reply);
				error (1, buf2);
				}
			fclose (tmpf1);
			rmlock (Newsdir, "nntp");
			goto quit;
			}

		nntp->newarticles = lines;
		}

	fclose (tmpf1);
	rmlock (Newsdir, "nntp");

	if (nntp->quit || main_exit)
		goto quit;

	if (nntpverbose && nntp->newarticles != 0)
		tprintf ("News available: %d articles\n", nntp->newarticles);

	if (nntpnewgroups)
		{

		/* Open the newgroups file */
		sprintf (buf1, "%s/newgroup", Newsdir);
		if (mlock (Newsdir, "newgroup"))
			{
			sprintf (buf2, "NNTP Can't lock file %s", buf1);
			error (2, buf2);
			goto quit;
			}

		if ((tmpf1 = fopen (buf1, APPEND_TEXT)) == NULLFILE)
			{
			sprintf (buf2, "NNTP Can't open file %s", buf1);
			error (1, buf2);
			rmlock (Newsdir, "newgroup");
			goto quit;
			}

		if (nntptrace >= 3)
			tprintf ("==>NEWGROUPS %s GMT\n", lastdate);

		usprintf (nntp->s, "NEWGROUPS %s GMT\n", lastdate);

		if ((reply = getreply ()) != 231 || (lines = gettxt (tmpf1, NULL, 0)) == -1)
			{
			if (reply == -1 || lines == -1)
				error (1, "NNTP Connection closed while getting NEWGROUPS");
			else
				{
				sprintf (buf2, "NNTP Error while getting NEWGROUPS (response was %d)", reply);
				error (1, buf2);
				}
			fclose (tmpf1);
			rmlock (Newsdir, "newgroup");
			goto quit;
			}

		fflush (tmpf1);
		fseek (tmpf1, 0L, SEEK_END);
		tmpfpos = ftell (tmpf1);
		fclose (tmpf1);
		if (tmpfpos == 0L)
			{
			sprintf (buf1, "%s/newgroup", Newsdir);
			remove (buf1);
			}

		rmlock (Newsdir, "newgroup");

		nntp->groupsreceived = lines;
		if (nntpverbose && nntp->groupsreceived != 0)
			tprintf ("News available: %d new groups\n", nntp->groupsreceived);

		}

	/* If it exists, open the get file */

	sprintf (buf1, "%s/get", Newsdir);
	if (access (buf1, 0) == 0)
		{
		if (mlock (Newsdir, "get"))
			{
			sprintf (buf2, "NNTP Can't lock file %s", buf1);
			error (2, buf2);
			}
		else
			{
			if ((nntp->getmsgf = fopen (buf1, READ_TEXT)) == NULLFILE)
				{
				sprintf (buf2, "NNTP Can't open file %s", buf1);
				error (1, buf2);
				rmlock (Newsdir, "get");
				}
			else
				{
				/* Open temporary file for unavailable message ids */
				sprintf (buf1, "%s/get.tmp", Newsdir);
				if ((gettmpf = fopen (buf1, WRITE_TEXT)) == NULLFILE)
					{
					sprintf (buf2, "NNTP Can't open file %s", buf1);
					error (1, buf2);
					fclose (nntp->getmsgf);
					nntp->getmsgf = NULLFILE;
					rmlock (Newsdir, "get");
					}

				}
			}
		}

#if NEWSBATCH
	/* Open the batch file */

	sprintf (buf1, "%s/batch.txt", News_spool ? News_spool : Mailspool);
	if (mlock (News_spool ? News_spool : Mailspool, "batch"))
		{
		sprintf (buf2, "NNTP Can't lock file %s", buf1);
		error (2, buf2);
		goto quit;
		}
	if ((msgf = fopen (buf1, RW_LOOKUP_TEXT)) == NULLFILE &&

		(msgf = fopen (buf1, RW_CREATE_TEXT)) == NULLFILE)
		{
		sprintf (buf2, "NNTP Can't open file %s", buf1);
		error (1, buf2);
		rmlock (News_spool ? News_spool : Mailspool, "batch");
		goto quit;
		}

	fseek (msgf, 0L, SEEK_END);
	msgfpos = ftell (msgf);
	fprintf (msgf, NEWSBATCHFORMAT, 0UL);
#else
	/* Open temporary file for messages */

	if ((msgf = tmpfile ()) == NULLFILE)
		{
		error (1, "NNTP Can't open temporary file");
		goto quit;
		}
#endif

	/* wait for the tables to be built then start up the tx proc */

	if (nntp->childproc != NULLPROC)
		if (pwait (Curproc) != 0)
			goto quit;

	if (nntp->quit || main_exit)
		goto quit;

	nntp->childproc = newproc ("NNTP client send", 1024, nntp_send, 0, NULL, Curproc, 0);

	nntp->receivednewschars = nntp->receivedtotalchars;
	nntp->startarticles = msclock ();

	while (!nntp->quit && !main_exit)
		{

		if (nntp->requesthead == nntp->requesttail)
			{
			if (nntp->childproc == NULLPROC)
				{
				error (2, "NNTP Receive internal error");
				goto quit;
				}

			if (pwait (Curproc) != 0)
				goto quit;

			}

		if (nntp->requesttail->status != reqend)
			{

			reply = getreply_in_buf (buf1);		/*  GeK - buffer captures response  */
			switch (reply)
				{

				case 205:						 /* closing connection */
					if (nntp->requesttail->status != reqquit)
						{
						internal_receive_error (reply, buf1, nntp->requesttail);
						goto quit;
						}

					sprintf (buf1, "%s/nntp.dat", Newsdir);
					if (mlock (Newsdir, "nntp"))
						{
						sprintf (buf2, "NNTP Can't lock file %s", buf1);
						error (2, buf2);
						goto quit;
						}

					if ((tmpf1 = fopen (buf1, READ_TEXT)) == NULLFILE)
						{
						sprintf (buf2, "NNTP Can't open file %s", buf1);
						error (1, buf2);
						rmlock (Newsdir, "nntp");
						goto quit;
						}

					sprintf (buf1, "%s/nntp.tmp", Newsdir);
					if ((tmpf2 = fopen (buf1, WRITE_TEXT)) == NULLFILE)
						{
						sprintf (buf2, "NNTP Can't open file %s", buf1);
						error (1, buf2);
						fclose (tmpf1);
						rmlock (Newsdir, "nntp");
						goto quit;
						}

					sprintf (buf1, "%s %s\n", np->name, datenow);
					i = strlen (np->name);
					cp = buf2 + i;
					while (fgets (buf2, NNTPMAXLEN, tmpf1) != NULLCHAR)
						if (strnicmp (buf2, np->name, i) != 0 || !isspace (*cp))
							fputs (buf2, tmpf2);
						else
							{
							fputs (buf1, tmpf2);
							buf1[0] = '\0';
							}

					if (buf1[0] != '\0')
						fputs (buf1, tmpf2);

					fclose (tmpf1);
					sprintf (buf1, "%s/nntp.dat", Newsdir);
					sprintf (buf2, "%s/nntp.tmp", Newsdir);
					if (ferror (tmpf2))
						{
						fclose (tmpf2);
						remove (buf2);
						error (1, "Error writing new nntp.dat file");
						}
					else
						{
						fclose (tmpf2);
						remove (buf1);
						rename (buf2, buf1);
						}

					rmlock (Newsdir, "nntp");
					goto quit;

				case 220:						 /* header and body follows */
					if (nntp->requesttail->status != reqgetarticle &&
						nntp->requesttail->status != reqnewarticle)
						{
						/*  GeK - log error, then continue rather than quitting  */
						internal_receive_error (reply, buf1, nntp->requesttail);
						nntp->requesttail->status = reqempty;
						break;
						}

					nntp->requesttail->lines = 0;
					nntp->requesttail->chars = 0UL;
					if (gettxt (msgf, nntp->requesttail, 0) == -1)
						{
						/*  GeK - log details of failed article  */
						rip (nntp->requesttail->msgid);
						sprintf (buf2, "NNTP receive failed after: %s %s: %s", "ARTICLE", nntp->requesttail->msgid, buf1);
						error (2, buf2);
						goto quit;
						}

					if (putarticle (msgf, &msgfpos, nntp->requesttail) == -1)
						goto quit;

					/* The following is a compromise: if there is a hash
					 * collision then the id will erroneously not get
					 * added. However, scanning the history file here will
					 * cause problems as the child might be doing the same. */

					if (nntp->requesttail->status == reqgetarticle)
						{
						if (!isinhashtable (nntp->requesttail->msgid))
							addtohistory (nntp->requesttail->msgid);

						nntp->getreceived++;
						}
					else
						{
						addtohistory (nntp->requesttail->msgid);
						nntp->newreceived++;
						}

					/*  GeK - compensate for chars not including \r  */
					nntp->receivedarticlechars += nntp->requesttail->chars + nntp->requesttail->lines;
					nntp->requesttail->status = reqempty;
					break;

				case 221:						 /* header follows */
					if (nntp->requesttail->status != reqheader)
						{
						/*  GeK - log error, then continue rather than quitting  */
						internal_receive_error (reply, buf1, nntp->requesttail);
						nntp->requesttail->status = reqempty;
						break;
						}

					nntp->requesttail->lines = 0;
					nntp->requesttail->chars = 0UL;
					if (getheader (nntp->requesttail) == -1)
						{
						/*  GeK - log details of failed header  */
						rip (nntp->requesttail->msgid);
						sprintf (buf2, "NNTP receive failed after: %s %s: %s", "HEAD", nntp->requesttail->msgid, buf1);
						error (2, buf2);
						goto quit;
						}

					if (nntp->requesttail->status == reqkill)
						{
						moveheadertofile (msgf, nntp->requesttail);
						if (putarticle (msgf, &msgfpos, nntp->requesttail) == -1)
							goto quit;

						addtohistory (nntp->requesttail->msgid);
						nntp->newheadersreceived++;
						/*  GeK - compensate for chars not including \r  */
						nntp->receivedarticlechars += nntp->requesttail->chars + nntp->requesttail->lines;
						nntp->requesttail->status = reqempty;
						}

					break;

				case 222:						 /* body follows */
					if (nntp->requesttail->status != reqbody)
						{
						/*  GeK - log error, then continue rather than quitting  */
						internal_receive_error (reply, buf1, nntp->requesttail);
						nntp->requesttail->status = reqempty;
						break;
						}

					moveheadertofile (msgf, nntp->requesttail);
					fputc ('\n', msgf);
					nntp->requesttail->lines++;
					nntp->requesttail->chars++;
					if (gettxt (msgf, nntp->requesttail, 0) == -1)
						{
						/*  GeK - log details of failed body  */
						rip (nntp->requesttail->msgid);
						sprintf (buf2, "NNTP receive failed after: %s %s: %s", "BODY", nntp->requesttail->msgid, buf1);
						error (2, buf2);
						goto quit;
						}

					if (putarticle (msgf, &msgfpos, nntp->requesttail) == -1)
						goto quit;

					addtohistory (nntp->requesttail->msgid);
					nntp->newreceived++;
					/*  GeK - compensate for chars not including \r  */
					nntp->receivedarticlechars += nntp->requesttail->chars + nntp->requesttail->lines;
					nntp->requesttail->status = reqempty;
					break;

				case 430:						 /* no such article */
				case 502:						 /*  GeK - no permission for article/head/body  */
				if (nntp->requesttail->status != reqgetarticle &&
						nntp->requesttail->status != reqnewarticle &&
						nntp->requesttail->status != reqheader &&
						nntp->requesttail->status != reqbody)
						{
						/*  GeK - log error, must be badly out of sync here so quit  */
						internal_receive_error (reply, buf1, nntp->requesttail);
						goto quit;
						}

					if (nntp->requesttail->status == reqgetarticle)
						{
						fputs (nntp->requesttail->msgid, gettmpf);
						nntp->getunavailable++;
						}
					else if (nntp->requesttail->status == reqbody)
						{
						/*
						 *  GeK - body could be unavailable even if header isn't,
						 *  act as if body killed ...
						 */
						moveheadertofile (msgf, nntp->requesttail);
						if (putarticle (msgf, &msgfpos, nntp->requesttail) == -1)
							goto quit;

						addtohistory (nntp->requesttail->msgid);
						nntp->newheadersreceived++;
						nntp->receivedarticlechars += nntp->requesttail->chars + nntp->requesttail->lines;
						}
					else
						nntp->newunavailable++;

#ifdef	notdef
					if (nntpverbose)
						tprintf ("News unavailable: article %s", nntp->requesttail->msgid);
#endif
					/*  GeK - add msg-id to error here if none in response, buf1 has no \n  */
					rip (nntp->requesttail->msgid);
					i = (strchr (buf1, '<') != NULL);
					sprintf (buf2, "NNTP %s unavailable%s%s: %s",
					         (nntp->requesttail->status == reqheader) ? "header" : ((nntp->requesttail->status == reqbody) ? "body" : "article"),
					         i ? "" : " ",
					         i ? "" : nntp->requesttail->msgid,
					         buf1);
					error (2, buf2);

					nntp->requesttail->status = reqempty;
					break;

#ifdef	notdef
				case -1:						 /* error */
#endif
				default:
					internal_receive_error (reply, buf1, nntp->requesttail);
					goto quit;

				}
			}

		/* move on to next request */

		nntp->requesttail = nntp->requesttail->nextrequest;
		psignal (nntp->childproc, 0);
		}

quit:

	nntp->endsession = msclock ();
	nntp->receivednewschars = nntp->receivedtotalchars - nntp->receivednewschars;

	/* tidy up as necessary */

#if NEWSBATCH
	if (msgf != NULLFILE)
		{
		fflush (msgf);
		chsize (fileno (msgf), msgfpos);
		fclose (msgf);
		if (msgfpos == 0L)
			{
			sprintf (buf1, "%s/batch.txt", News_spool ? News_spool : Mailspool);
			(void) remove (buf1);
			}

		rmlock (News_spool ? News_spool : Mailspool, "batch");
		}
#else
	if (msgf != NULLFILE)
		{
		fclose (msgf);
		}
#endif

	nntp->quit = TRUE;
	if (nntp->childproc != NULLPROC)
		pwait (NULL);

	if (nntp->childproc != NULLPROC)
		killproc (nntp->childproc);

	/* Update get file with message ids we didn't get */

	if (nntp->getmsgf != NULLFILE)
		{
		rewind (nntp->getmsgf);
		for (i = 0; i < (nntp->getinvalid + nntp->getunavailable + nntp->getreceived); i++)
			fgets (buf1, NNTPMAXLEN, nntp->getmsgf);

		while (fgets (buf1, NNTPMAXLEN, nntp->getmsgf) != NULLCHAR)
			fputs (buf1, gettmpf);

		fclose (nntp->getmsgf);
		sprintf (buf1, "%s/get", Newsdir);
		sprintf (buf2, "%s/get.tmp", Newsdir);
		if (ferror (gettmpf))
			{
			fclose (gettmpf);
			remove (buf2);
			error (1, "Error writing new get file");
			}
		else
			{
			remove (buf1);
			fflush (gettmpf);
			fseek (gettmpf, 0L, SEEK_END);
			tmpfpos = ftell (gettmpf);
			fclose (gettmpf);
			if (tmpfpos == 0L)
				remove (buf2);
			else
				rename (buf2, buf1);
			}

		rmlock (Newsdir, "get");
		}

	if (nntp->newmsgf != NULLFILE)
		fclose (nntp->newmsgf);

	if (nntp->historyf != NULLFILE)
		{
		fclose (nntp->historyf);
		rmlock (Newsdir, "history");
		}

	if (nntp->hashtable != NULL)
		free (nntp->hashtable);

	/* Free all dynamically allocated buffer space */

		{
		struct nntprequest
		           *tail = nntp->requesthead, *head = tail->nextrequest;

		tail->nextrequest = NULLNNTPREQUEST;
		while (head != NULLNNTPREQUEST)
			{
			freetextlinelist (&head->header);
			free (head->newsgroup_list);	/*  GeK  */
			tail = head;
			head = head->nextrequest;
			free (tail);
			}

		for (i = 0; i < 27; i++)	/*  GeK - free each list  */
			{
			freetextlinelist (&nntp->killlist[i]);
			freetextlinelist (&nntp->keeplist[i]);
			}
		}

	/* Output statistics */
	if (nntp->getreceived != 0 || nntp->newheadersreceived != 0 || nntp->newreceived != 0)
		{
		sprintf (buf2, "News summary: %u articles (%lu bytes) in %lu sec (%lu bytes/sec)",
		   nntp->getreceived + nntp->newheadersreceived + nntp->newreceived,
				 nntp->receivedarticlechars,
				 (nntp->endsession - nntp->startarticles) / 1000UL,
				 unitspersecond (nntp->receivedarticlechars, nntp->endsession - nntp->startarticles));
		log (nntp->s, buf2);
		tprintf (nntpverbose ? "\n%s\n" : "%s\n", buf2);
		}

	if (nntp->getarticles != 0)
		{
		sprintf (buf2, "Get articles: %u invalid, %u unavailable, %u received",
				 nntp->getinvalid,
				 nntp->getunavailable,
				 nntp->getreceived);
		log (nntp->s, buf2);
		if (nntpverbose)
			tprintf ("%s\n", buf2);

		}

	if (nntp->newarticles != 0)
		{
		sprintf (buf2, "New articles: %u duplicate, %u unavailable, %u headers, %u complete",
				 nntp->newduplicates,
				 nntp->newunavailable,
				 nntp->newheadersreceived,
				 nntp->newreceived);
		log (nntp->s, buf2);
		if (nntpverbose)
			tprintf ("%s\n", buf2);

		}

	if (nntp->getreceived != 0 || nntp->newheadersreceived != 0 || nntp->newreceived != 0 ||
		nntp->getarticles != 0 || nntp->newarticles != 0)
		{
		sprintf (buf2, "History file: %u entries, %u complete scans",
				 nntp->historyentries,
				 nntp->historyscans);
		log (nntp->s, buf2);
		if (nntpverbose)
			tprintf ("%s\n", buf2);
		sprintf (buf2, "Throughput  : %lu/%lu bytes in %lu/%lu sec (%lu/%lu bytes/sec)",
				 nntp->receivednewschars,
				 nntp->receivedtotalchars,
				 (nntp->endsession - nntp->startarticles) / 1000UL,
				 (nntp->endsession - nntp->startsession) / 1000UL,
				 unitspersecond (nntp->receivednewschars, nntp->endsession - nntp->startarticles),

				 unitspersecond (nntp->receivedtotalchars, nntp->endsession - nntp->startsession));
		log (nntp->s, buf2);
		if (nntpverbose)
			tprintf ("%s\n\n", buf2);

		}

	if (nntpverbose)
		tprintf ("Closing news session\n");

	/*  GeK - log close event  */
	log (nntp->s, "NNTP close %s", np->name);

	close_s (nntp->s);
	free (nntp);
	nntp = NULLNNTPSESSION;

	if (nntptrace >= 3)
		tprintf ("NNTP daemon exiting\n");

	/* Restart timer */

	start_timer (&np->nntpcli_t);

#ifdef	NNTP_SESSION
	sp->s = -1;
	alarm (120000L);	/*  2 minute timeout  */
	keywait (NULL, 1);
	alarm (0L);
	freesession (sp);
#endif	/* NNTP_SESSION */

	return;
	}

#ifdef	NNTP_POST
/*
 *	Direct posting support for NNTP client.
 *
 *	This code started life as separate file with 'nntp_post()' taking
 *	a 'int socket' parameter.  However, we want to call several funcs
 *	in 'nntpcli' which are defined there as static and which in turn
 *	reference the static 'nntp->s' socket index, already set up and
 *	ready to use at the time 'nntp_post()' is invoked.
 *
 *	Perhaps this file should just reside in nntpcli.c (subject to a
 *	#define of course) as it is arguably part of the client, but for
 *	now it is just #include'd in there, after 'nntp_main()'.
 *
 *	PDL Note:
 *	The strdup() here is that defined in misc.c and uses mallocw().
 *	Do we therefore need to set 'alarm()'s ???
 *  [I'd say no; it doesn't bother when doing mallocw() elsewhere. -SCT]
 *
 *	PDL Note 2:
 *	These comments need sorting out before this source goes public.
 */


/*
 * PDL:
 * No idea which 'nntp post ...' commands were envisaged, so this is
 * just a stub for now.  Connection, socket, newsserver etc. is already
 * established so nothing to set up in that dept.  Some thoughts:
 *
 *	nntp post [on | off]		(override 'nntp_posting_allowed'?)
 *	nntp post extn "nws"
 *
 * SCT:
 * For the moment, just a single command, to enable/disable NNTP posting,
 * and allow an "NNTP posting only" mode, where NNTP posting is done but
 * not the usual news grabbing:
 *
 *	nntp post [off | on | only]
 */
int do_nntp_post (int argc, char *argv[], void *p)
{
	if (argc < 2)
	{
		tprintf ("NNTP posting %s\n", (nntppost > 1) ? "only" :
			(nntppost > 0) ? "on" : "off");
		return 0;
	}
	if (strcmpi(argv[1], "off") == 0)
		nntppost = 0;
	else if (strcmpi(argv[1], "on") == 0)
		nntppost = 1;
	else if (strcmpi(argv[1], "only") == 0)
		nntppost = 2;
	else
		tprintf ("Valid options: off on only\n");
	return 0;
}

/*
 * This is the main routine for posting articles - it looks in NNTPpostdir
 * for any "*.txt" (was "*.nws") files and attempts to POST them.
 * Returns -1 on error, else 0.
 */
int nntp_post (void)
{
FILE	*fp;
struct	ffblk  fb;
char	fullpath [MAXPATH];
char	*filepart;
char	*msgid;
int retcode = 0;

	sprintf (fullpath, "%s/", NNTPpostdir);
	filepart = fullpath + strlen (fullpath);
	strcpy (filepart, "*.txt");
	if (findfirst (fullpath, &fb, FA_NORMAL) == 0)
	{
		do
		{
			strcpy (filepart, fb.ff_name);
			strlwr (filepart);
			msgid = NULLCHAR;

			/* do we need to lock this file?? */

			if ((fp = validate_article (fullpath, &msgid)) == NULLFILE)
				continue;

			retcode = post_article (fp, msgid, filepart);
			fclose (fp);
			free (msgid);

			if (retcode < 0)	/* some sort of fatal error */
				break;

			/* let someone else do their stuff */
			if (!nntp->quit && !main_exit)
				pwait (NULL);

			if (retcode == 0)	/* posted successfully */
				unlink (fullpath);
		}
		while (!nntp->quit && !main_exit && findnext (&fb) == 0);
	}

	/*
	 * If we're posting only, issue a "QUIT" here and get the reply before
	 * we return from this function.
	 */
	if (retcode == 0 && nntppost > 1)	/* posting only: issue "QUIT" now */
	{
		char *cp = "==>QUIT\n";
		if (nntptrace >= 3)
			tputs (cp);
		usprintf (nntp->s, cp + 3);
		alarm (nntppost_respwait);
		getreply();			/* Don't bother checking code: must be 205 */
		alarm (0L);
	}

	return retcode;
}

/*
 * Check that the article is syntactically OK before attempting to post it.
 * All we do here is to check that all the mandatory headers are present (and
 * extract the msg-ID while so doing).	Checking is pretty rudimentary, but
 * arguably that's the job of the newsreader that created the article...
 * However, we don't want to look stupid to the news server.
 *
 * If all OK, initialise the msg-ID and return a FILE ptr to the open file.
 */

#define H_PATH	0x01	/* bit/flag cleared when valid "Path:" found */
#define H_FROM	0x02	/* etc. */
#define H_DATE	0x04
#define H_SUBJ	0x08
#define H_NGRP	0x10
#define H_MSID	0x20

#define H_MANDATORY (H_PATH | H_FROM | H_DATE | H_SUBJ | H_NGRP | H_MSID)

static
FILE	*validate_article (const char *pathname, char **pmsgid)
{
FILE	*fp;
char	*cp;
char	*msgid;
int last_had_newline;	/*GeK - prevent breaking long lines */
int needhdr;
char	buffer [NNTPMAXLEN];

	if ((fp = fopen (pathname, READ_TEXT)) == NULLFILE)
	{
		/* buffer should be more than big enough to store this */
		sprintf (buffer, "NNTP post: Can't access file `%s'", pathname);
		error (2, buffer);
		return NULLFILE;
	}

	needhdr = H_MANDATORY;		/* we need all these headers */
	msgid = NULLCHAR;
	last_had_newline = TRUE;
	while (fgets (buffer, sizeof (buffer), fp) != NULLCHAR)
	{
		if (last_had_newline)
		{
			/* only bother to check the header lines */
			if (*buffer == '\n')
				break;		/* end of header */

			/* check the 6 mandatory (rfc1036) headers */
			if ((needhdr & H_PATH) && strnicmp (buffer, "Path:", 5) == 0)
			{
				cp = skipws (buffer + 5);
				if (strchr (cp, '.') != NULLCHAR)       /* ??? */
					needhdr &= ~H_PATH;
			}
			else if ((needhdr & H_FROM) && strnicmp (buffer, "From:", 5) == 0)
			{
				cp = skipws (buffer + 5);
				if (strchr (cp, '@') != NULLCHAR)       /* ??? */
					needhdr &= ~H_FROM;
			}
			else if ((needhdr & H_DATE) && strnicmp (buffer, "Date:", 5) == 0)
			{
				cp = skipws (buffer + 5);	/* min: "dd mmm yy hh:mm x" */
				if (strlen (cp) > 17 && strchr (cp, ':') != NULLCHAR)
					needhdr &= ~H_DATE;
			}
			else if ((needhdr & H_SUBJ) && strnicmp (buffer, "Subject:", 8) == 0)
			{
				cp = skipws (buffer + 8);
				if (strlen (cp) > 1)			/* ??? */
					needhdr &= ~H_SUBJ;
			}
			else if ((needhdr & H_NGRP) && strnicmp (buffer, "Newsgroups:", 11) == 0)
			{
				cp = skipws (buffer + 11);
				if (strlen (cp) > 3 && strchr (cp, '.') != NULLCHAR)
					needhdr &= ~H_NGRP;
			}
			else if ((needhdr & H_MSID) && strnicmp (buffer, "Message-ID:", 11) == 0)
			{
				char	*ap;	/* '@' ptr */
				char	*ep;	/* end ptr */

				cp = skipws (buffer + 11);
				if (*cp == '<'
				&& (ap = strchr (cp + 1, '@')) != NULLCHAR
				&& (ep = strchr (ap + 1, '>')) != NULLCHAR)
				{
					/* allow for '<', '>', "\n\0" */
					if (ep - cp < MAX_MESSAGE_ID_LEN - 3)
					{
						*ep = '\0';		/* hide '>' onwards for strdup */
						msgid = strdup (cp + 1);
						*ep = '>';		/* restore rest of line */
						needhdr &= ~H_MSID;
					}
				}
			}
		}
		last_had_newline = (strchr (buffer, '\n') != NULL);
	}

	/* check that we got them all */
	if (!needhdr)
	{
		*pmsgid = msgid;	/* all OK -- return */
		return (fp);
	}

	/*
	 * Else something wrong: we could be more specific on which headers
	 * are wrong/missing but for now we'll be general.
	 */
	fclose (fp);
	if (msgid != NULLCHAR)
		free (msgid);
	cp = strrchr (pathname, '/');   /* guaranteed... */
	sprintf (buffer,
		"NNTP post: Missing or malformed header(s) in article `%s' (0x%02x)",
		cp + 1, needhdr);
	error (0, buffer);
	return NULLFILE;
}

#undef	H_PATH
#undef	H_FROM
#undef	H_DATE
#undef	H_SUBJ
#undef	H_NGRP
#undef	H_MSID
#undef	H_MANDATORY

/*
 * Finally post the article.  Or try to...
 * Note: getreply() skips empty lines and 1xx info messages.
 */
static
int post_article (FILE *fp, const char *msgid, const char *fname)
{
int response;
char	*cp;
char	*jobno;
char	buffer [NNTPMAXLEN];

	/* extract the job number from the filename */
	jobno = strdup (fname);
	if ((cp = strchr (jobno, '.')) != NULLCHAR) /* should be! */
		*cp = '\0';

	cp = "==>POST\n";
	if (nntptrace >= 3)
		tputs (cp);
	usputs (nntp->s, cp + 3);

	alarm (nntppost_respwait);
	response = getreply();
	alarm (0L);
	if (response == 340)				/* OK - send article */
	{
		int  last_had_newline = TRUE;	/* prevent breaking long lines */

		if (nntpverbose || nntptrace >= 1)
			tprintf ("NNTP post: Sending job %s msg-id <%s>\n", jobno, msgid);

		rewind (fp);
		while (fgets (buffer, sizeof (buffer), fp) != NULLCHAR)
		{	/*
			 * Dot transparency: add another '.' at the start of
			 * any line which already starts with a '.'
			 */
			if (last_had_newline && *buffer == '.')
			{
				if (nntptrace >= 4)
					tprintf ("==>.%s", buffer);
				usputc (nntp->s, '.');
			}
			else if (nntptrace >= 4)
				tprintf ("==>%s", buffer);

			usputs (nntp->s, buffer);
			last_had_newline = (strchr (buffer, '\n') != NULL);
		}

		/* Send end-of-message command */
		if (!last_had_newline)
		{
			if (nntptrace >= 4)
				tputs ("==>\n");
			usputc (nntp->s, '\n');
		}

		cp = "==>.\n";
		if (nntptrace >= 3)
			tputs (cp);
		usputs (nntp->s, cp + 3);

		if (nntptrace >= 2)
			tputs ("NNTP post: Sent \"\\n.\\n\"\n");

		alarm (nntppost_respwait);
		response = getreply();	/* will log any error [r >= 400] */
		alarm (0L);

		sprintf (buffer, "NNTP post: Job %s msg-id <%s> %s (%d)", jobno, msgid,
			(response == 240) ? "accepted" : "rejected", response);
		log (nntp->s, "%s", buffer);
		if (nntptrace >= 2)						/* display long message */
			tprintf ("%s\n", buffer);
		else if (nntpverbose || nntptrace == 1)	/* display shorter message */
		{
			if (response == 240)
				tprintf ("NNTP post: Job %s accepted\n", jobno);
			else
				tprintf ("NNTP post: Job %s rejected after posting (%d)\n",
					jobno, response);
		}
	}
	else								/* Response != 340; can't post */
	{
		sprintf (buffer, "NNTP post: Job %s: POST command refused (%d)",
			jobno, response);
		error (1, buffer);
	}

	free (jobno);

	/*
	 * 240 - posted OK
	 * 440 - posting prohibited
	 * 441 - posting failed
	 */
	return (response == 240) ? 0 :
	       (response == 441) ? 1 : -1 ;
}

static
char	*skipws (const char *s)
{
	while (*s == ' ' || *s == '\t')
		++s;

	return (char *) s;
}
#endif	/* NNTP_POST */

/*  GeK - more detailed logging of unexpected state machine errors  */
static void
internal_receive_error (int response, char *response_buf, struct nntprequest *request)
{
	static const char *state[] =
	{
		"empty",		 /* vacant request */
		"getart",		 /* article from get file */
		"newart",		 /* article from new file */
		"header",		 /* header only */
		"body",			 /* body to accompany header */
		"killhdr",		 /* kill this header */
		"keephdr",		 /* get a body to accompany this header */
		"end",			 /* no more requests */
		"quit"			 /* close connection */
	};
	char   *buf;
	int     has_id;

	/*  Assume IDs for getart, newart, header, body.  Ick.  I know, but it's compact  */
	has_id = (request->status >= reqgetarticle && request->status <= reqbody);
	if (has_id)
		rip (request->msgid);

	/*  Assumes response_buf has no \n  */
	buf = malloc (64 + strlen (response_buf) + (has_id ? strlen (request->msgid) : 0));
	if (buf != NULL)
	{
		sprintf (buf, "NNTP %s in state %u [%s] %s: %s",
		         (response == -1) ? "Connection closed" : "Unexpected response",
		   	     request->status, state [request->status],
		         has_id ? request->msgid : "",
		         response_buf);
		error (1, buf);
		free (buf);
	}
	else
	{
		/*  Memory tight  */
		error (1, "NNTP receive error");
	}
}

#ifdef	notdef	/*  GeK - removed this  */
static void nntp_reg (void)
	{
	struct sockaddr_in sock;
	struct socket lsocket;
	struct socket rsocket;
	struct usock *usp;
	struct sockaddr_in *sinp;
	struct mbuf *bp;
	char buf[64];
	int s;

	sock.sin_family = AF_INET;
	sock.sin_port = 2911;
	if ((sock.sin_addr.s_addr = resolve ("passwd.demon.co.uk")) == 0 ||
		(s = socket (AF_INET, SOCK_DGRAM, 0)) == -1 ||
		connect (s, (char *) &sock, sizeof (sock)) == -1)
		return;

	(void) sprintf (buf, "%s|%s|%s\n", Version, __DATE__, __TIME__);
	bp = ambufw (strlen (buf) + 1);
	(void) strcpy (bp->data, buf);
	usp = itop (s);
	sinp = (struct sockaddr_in *) usp->name;
	lsocket.address = sinp->sin_addr.s_addr;
	lsocket.port = sinp->sin_port;
	sinp = (struct sockaddr_in *) usp->peername;
	rsocket.address = sinp->sin_addr.s_addr;
	(void) send_udp (&lsocket, &rsocket, 0, 64, bp, bp->size, 0, 0);
	close_s (s);
	}						 /* static void nntp_reg (void) */
#endif	/* notdef */
